"""Base class for embedders and models."""

import abc
from pathlib import Path
from typing import Any, Dict, TypeVar, Union

import geopandas as gpd
import pandas as pd

from srai.constants import GEOMETRY_COLUMN

try:  # pragma: no cover
    from pytorch_lightning import LightningModule

except ImportError:
    from srai.embedders._pytorch_stubs import LightningModule


class Model(LightningModule):  # type: ignore
    """Class for model based on LightningModule."""

    def get_config(self) -> Dict[str, Any]:
        """Get model config."""
        model_config = {
            k: v
            for k, v in vars(self).items()
            if k[0] != "_"
            and k
            not in (
                "training",
                "prepare_data_per_node",
                "allow_zero_length_dataloader_with_multiple_devices",
            )
        }

        return model_config

    def save(self, path: Union[Path, str]) -> None:
        """
        Save the model to a directory.

        Args:
            path (Path): Path to the directory.
        """
        import torch

        torch.save(self.state_dict(), path)

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs: Any) -> "Model":
        """
        Load model from a file.

        Args:
            path (Union[Path, str]): Path to the file.
            **kwargs (dict): Additional kwargs to pass to the model constructor.
        """
        import torch

        if isinstance(path, str):
            path = Path(path)

        model = cls(**kwargs)
        model.load_state_dict(torch.load(path))
        return model


class Embedder(abc.ABC):
    """Abstract class for embedders."""

    @abc.abstractmethod
    def transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> pd.DataFrame:  # pragma: no cover
        """
        Embed regions using features.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.

        Returns:
            pd.DataFrame: Embedding and geometry index for each region in regions_gdf.

        Raises:
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
        """
        raise NotImplementedError

    # @abc.abstractmethod
    # def save(self, path: Union[Path, str]) -> None:
    #     """
    #     Save the embedder to a directory.

    #     Args:
    #         path (Path): Path to the directory.
    #     """
    #     raise NotImplementedError

    # @classmethod
    # @abc.abstractmethod
    # def load(cls, path: Union[Path, str]) -> "Embedder":
    #     """
    #     Load the embedder from a directory.

    #     Args:
    #         path (Path): Path to the directory.

    #     Returns:
    #         Embedder: The loaded embedder.
    #     """
    #     raise NotImplementedError

    def _validate_indexes(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> None:
        if regions_gdf.index.name is None:
            raise ValueError("regions_gdf must have a named index.")

        if features_gdf.index.name is None:
            raise ValueError("features_gdf must have a named index.")

        if not isinstance(joint_gdf.index, pd.MultiIndex):
            raise ValueError(
                f"joint_gdf.index must be of type pandas.MultiIndex, not {type(joint_gdf.index)}"
            )

        if len(joint_gdf.index.names) != 2:
            raise ValueError(
                f"joint_gdf.index must have 2 levels, has {len(joint_gdf.index.names)}"
            )

        if regions_gdf.index.name != joint_gdf.index.names[0]:
            raise ValueError(
                f"Name of regions_gdf.index ({regions_gdf.index.name}) must be equal to the name of"
                f" the 1st level of joint_gdf.index ({joint_gdf.index.names[0]})"
            )

        if features_gdf.index.name != joint_gdf.index.names[1]:
            raise ValueError(
                f"Name of features_gdf.index ({features_gdf.index.name}) must be equal to the name"
                f" of the 2nd level of joint_gdf.index ({joint_gdf.index.names[1]})"
            )

    def _remove_geometry_if_present(self, data: gpd.GeoDataFrame) -> pd.DataFrame:
        if GEOMETRY_COLUMN in data.columns:
            data = data.drop(columns=GEOMETRY_COLUMN)
        return pd.DataFrame(data)


ModelT = TypeVar("ModelT", bound=Model)



"""
NeighbourDataset.

This dataset is used to train a model to predict whether regions are neighbours or not.
As defined in Hex2Vec paper[1].

References:
    [1] https://dl.acm.org/doi/10.1145/3486635.3491076
"""
from typing import TYPE_CHECKING, Any, Dict, Generic, List, NamedTuple, Set, TypeVar

import numpy as np
import pandas as pd
from tqdm import tqdm

from srai._optional import import_optional_dependencies
from srai.neighbourhoods import Neighbourhood

if TYPE_CHECKING:  # pragma: no cover
    import torch

try:  # pragma: no cover
    from torch.utils.data import Dataset

except ImportError:
    from srai.embedders._pytorch_stubs import Dataset


T = TypeVar("T")


class NeighbourDatasetItem(NamedTuple):
    """
    Neighbour dataset item.

    Attributes:
        X_anchor (torch.Tensor): Anchor regions.
        X_positive (torch.Tensor): Positive regions.
            Data for the regions that are neighbours of regions in X_anchor.
        X_negative (torch.Tensor): Negative regions.
            Data for the regions that are NOT neighbours of the regions in X_anchor.
    """

    X_anchor: "torch.Tensor"
    X_positive: "torch.Tensor"
    X_negative: "torch.Tensor"


class NeighbourDataset(Dataset[NeighbourDatasetItem], Generic[T]):  # type: ignore
    """
    Dataset for training a model to predict neighbours.

    It works by returning triplets of regions: anchor, positive and negative. A model can be trained
    to predict that the anchor region is a neighbour of the positive region, and that it is not a
    neighbour of the negative region.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        neighbourhood: Neighbourhood[T],
        negative_sample_k_distance: int = 2,
    ):
        """
        Initialize NeighbourDataset.

        Args:
            data (pd.DataFrame): Data to use for training. Raw counts of features in regions.
            neighbourhood (Neighbourhood[T]): Neighbourhood to use for training.
                It has to be initialized with the same data as the data argument.
            negative_sample_k_distance (int): How many neighbours away to sample negative regions.
                For example, if k=2, then the negative regions will be sampled from regions that are
                at least 3 hops away from the anchor region. Has to be >= 2.

        Raises:
            ValueError: If negative_sample_k_distance < 2.
        """
        import_optional_dependencies(dependency_group="torch", modules=["torch"])
        import torch

        self._data = torch.Tensor(data.to_numpy())
        self._assert_negative_sample_k_distance_correct(negative_sample_k_distance)
        self._negative_sample_k_distance = negative_sample_k_distance

        self._anchor_df_locs_lookup: np.ndarray
        self._positive_df_locs_lookup: np.ndarray
        self._excluded_from_negatives: Dict[int, Set[int]] = {}

        self._region_index_to_df_loc: Dict[T, int] = {
            region_index: i for i, region_index in enumerate(data.index)
        }
        self._df_loc_to_region_index: Dict[int, T] = {
            i: region_index for region_index, i in self._region_index_to_df_loc.items()
        }

        self._build_lookup_tables(data, neighbourhood)

    def _build_lookup_tables(self, data: pd.DataFrame, neighbourhood: Neighbourhood[T]) -> None:
        anchor_df_locs_lookup: List[int] = []
        positive_df_locs_lookup: List[int] = []

        for region_df_loc, region_index in tqdm(enumerate(data.index), total=len(data)):
            region_direct_neighbours = neighbourhood.get_neighbours(region_index)
            neighbours_df_locs: Set[int] = {
                self._region_index_to_df_loc[neighbour_index]
                for neighbour_index in region_direct_neighbours
            }
            anchor_df_locs_lookup.extend([region_df_loc] * len(neighbours_df_locs))
            positive_df_locs_lookup.extend(neighbours_df_locs)

            indices_excluded_from_negatives = neighbourhood.get_neighbours_up_to_distance(
                region_index, self._negative_sample_k_distance
            )
            self._excluded_from_negatives[region_df_loc] = {
                self._region_index_to_df_loc[excluded_index]
                for excluded_index in indices_excluded_from_negatives
            }

        self._anchor_df_locs_lookup = np.array(anchor_df_locs_lookup)
        self._positive_df_locs_lookup = np.array(positive_df_locs_lookup)

    def __len__(self) -> int:
        """
        Return the number of anchor-positive pairs available in the dataset.

        Returns:
            int: The number of pairs.
        """
        return len(self._anchor_df_locs_lookup)

    def __getitem__(self, data_row_index: Any) -> NeighbourDatasetItem:
        """
        Return a single dataset item (anchor, positive, negative).

        Args:
            data_row_index (Any): The index of the dataset item to return.

        Returns:
            NeighbourDatasetItem: The dataset item.
                This includes the anchor region, positive region
                and arandomly sampled negative region.
        """
        anchor_df_loc = self._anchor_df_locs_lookup[data_row_index]
        positive_df_loc = self._positive_df_locs_lookup[data_row_index]
        negative_df_loc = self._get_random_negative_df_loc(anchor_df_loc)

        anchor_region = self._data[anchor_df_loc]
        positive_region = self._data[positive_df_loc]
        negative_region = self._data[negative_df_loc]

        return NeighbourDatasetItem(anchor_region, positive_region, negative_region)

    def _get_random_negative_df_loc(self, input_df_loc: int) -> int:
        excluded_df_locs = self._excluded_from_negatives[input_df_loc]
        negative_candidate: int = np.random.randint(0, len(self._data))  # noqa: NPY002
        while negative_candidate in excluded_df_locs:
            negative_candidate = np.random.randint(0, len(self._data))  # noqa: NPY002
        return negative_candidate

    def _assert_negative_sample_k_distance_correct(self, negative_sample_k_distance: int) -> None:
        if negative_sample_k_distance < 2:
            raise ValueError(
                "negative_sample_k_distance must be at least 2, "
                f"but was {negative_sample_k_distance}"
            )



"""
OSM popular tags downloader.

This module exposes a function `get_popular_tags` that uses taginfo[1] API
to download the most popular tags from OSM

References:
    1. https://taginfo.openstreetmap.org/
"""
from typing import Any, Dict, List

import requests
from functional import seq

from srai.loaders.osm_loaders.filters import OsmTagsFilter

_TAGINFO_API_ADDRESS = "https://taginfo.openstreetmap.org"
_TAGINFO_API_TAGS = _TAGINFO_API_ADDRESS + "/api/4/tags/popular"


def get_popular_tags(
    in_wiki_only: bool = False, min_count: int = 0, min_fraction: float = 0.0
) -> OsmTagsFilter:
    """
    Download the OSM's most popular tags from taginfo api.

    This is a wrapper around the `popular` taginfo api endpoint [1].
    It queries the API, and optionally filters the results
    according to argument values.

    Args:
        in_wiki_only (bool, optional): If True, only return results tags
            that have at least one wiki page. Defaults to False.
        min_count(int, optional): Minimum number of objects in OSM with this tag attached,
            to include the tag in the results. Defaults to 0 (no filtering).
        min_fraction(float, optional): What fraction of all objects have to have this tag attached,
            to include it in the results. Defaults to 0.0 (no filtering).

    Returns:
        Dict[str, List[str]]: dictionary containing the downloaded popular tags.

    References:
        1. https://taginfo.openstreetmap.org/taginfo/apidoc#api_4_tags_popular
    """
    taginfo_api_response = requests.get(
        _TAGINFO_API_TAGS,
        headers={"User-Agent": "SRAI Python package (https://github.com/kraina-ai/srai)"},
    )
    taginfo_api_response.raise_for_status()
    taginfo_data = taginfo_api_response.json()["data"]
    return _parse_taginfo_response(taginfo_data, in_wiki_only, min_count, min_fraction)


def _parse_taginfo_response(
    taginfo_data: List[Dict[str, Any]], in_wiki_only: bool, min_count: int, min_fraction: float
) -> OsmTagsFilter:
    result_tags = (
        seq(taginfo_data)
        .filter(lambda t: t["count_all"] >= min_count)
        .filter(lambda t: t["count_all_fraction"] >= min_fraction)
    )
    if in_wiki_only:
        result_tags = result_tags.filter(lambda t: t["in_wiki"])
    taginfo_grouped: OsmTagsFilter = (
        result_tags.map(lambda t: (t["key"], t["value"])).group_by_key().to_dict()
    )
    return taginfo_grouped



"""
Count Embedder.

This module contains count embedder implementation.
"""
from typing import List, Optional

import geopandas as gpd
import pandas as pd

from srai.embedders import Embedder


class CountEmbedder(Embedder):
    """Simple Embedder that counts occurences of feature values."""

    def __init__(
        self, expected_output_features: Optional[List[str]] = None, count_subcategories: bool = True
    ) -> None:
        """
        Init CountEmbedder.

        Args:
            expected_output_features (List[str], optional): The features that are expected
                to be found in the resulting embedding. If not None, the missing features are added
                and filled with 0. The unexpected features are removed.
                The resulting columns are sorted accordingly. Defaults to None.
            count_subcategories (bool, optional): Whether to count all subcategories individually
                or count features only on the highest level based on features column name.
                Defaults to True.
        """
        self.expected_output_features: Optional[pd.Series] = (
            None if expected_output_features is None else pd.Series(expected_output_features)
        )

        self.count_subcategories = count_subcategories

    def transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> pd.DataFrame:
        """
        Embed a given GeoDataFrame.

        Creates region embeddings by counting the frequencies of each feature value.
        Expects features_gdf to be in wide format with each column
        being a separate type of feature (e.g. amenity, leisure)
        and rows to hold values of these features for each object.
        The resulting DataFrame will have columns made by combining
        the feature name (column) and value (row) e.g. amenity_fuel or type_0.
        The rows will hold numbers of this type of feature in each region.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.

        Returns:
            pd.DataFrame: Embedding for each region in regions_gdf.

        Raises:
            ValueError: If features_gdf is empty and self.expected_output_features is not set.
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
        """
        self._validate_indexes(regions_gdf, features_gdf, joint_gdf)
        if features_gdf.empty:
            if self.expected_output_features is not None:
                return pd.DataFrame(
                    0, index=regions_gdf.index, columns=self.expected_output_features
                )
            else:
                raise ValueError(
                    "Cannot embed with empty features_gdf and no expected_output_features."
                )

        regions_df = self._remove_geometry_if_present(regions_gdf)
        features_df = self._remove_geometry_if_present(features_gdf)
        joint_df = self._remove_geometry_if_present(joint_gdf)

        if self.count_subcategories:
            feature_encodings = pd.get_dummies(features_df)
        else:
            feature_encodings = features_df.notna().astype(int)
        joint_with_encodings = joint_df.join(feature_encodings)
        region_embeddings = joint_with_encodings.groupby(level=0).sum()

        region_embeddings = self._maybe_filter_to_expected_features(region_embeddings)
        region_embedding_df = regions_df.join(region_embeddings, how="left").fillna(0).astype(int)

        return region_embedding_df

    def _maybe_filter_to_expected_features(self, region_embeddings: pd.DataFrame) -> pd.DataFrame:
        """
        Add missing and remove excessive columns from embeddings.

        Args:
            region_embeddings (pd.DataFrame): Counted frequencies of each feature value.

        Returns:
            pd.DataFrame: Embeddings with expected columns only.
        """
        if self.expected_output_features is None:
            return region_embeddings

        missing_features = self.expected_output_features[
            ~self.expected_output_features.isin(region_embeddings.columns)
        ]
        missing_features_df = pd.DataFrame(
            0, index=region_embeddings.index, columns=missing_features
        )
        region_embeddings = pd.concat([region_embeddings, missing_features_df], axis=1)
        region_embeddings = region_embeddings[self.expected_output_features]
        return region_embeddings



"""
OSM PBF Loader.

This module contains loader capable of loading OpenStreetMap features from `*.osm.pbf` files.
"""
from pathlib import Path
from typing import Hashable, List, Mapping, Optional, Sequence, Union

import geopandas as gpd
import pandas as pd

from srai._optional import import_optional_dependencies
from srai.constants import FEATURES_INDEX, GEOMETRY_COLUMN, WGS84_CRS
from srai.loaders.osm_loaders._base import OSMLoader
from srai.loaders.osm_loaders.filters import GroupedOsmTagsFilter, OsmTagsFilter
from srai.loaders.osm_loaders.pbf_file_downloader import PbfSourceLiteral


class OSMPbfLoader(OSMLoader):
    """
    OSMPbfLoader.

    OSM(OpenStreetMap)[1] PBF(Protocolbuffer Binary Format)[2] loader is a loader
    capable of loading OSM features from a PBF file. It filters features based on OSM tags[3]
    in form of key:value pairs, that are used by OSM users to give meaning to geometries.

    This loader uses `pyosmium`[3] library capable of parsing an `*.osm.pbf` file.

    Additionally, it can download a pbf file extract for a given area using different sources.

    References:
        1. https://www.openstreetmap.org/
        2. https://wiki.openstreetmap.org/wiki/PBF_Format
        3. https://osmcode.org/pyosmium/
    """

    def __init__(
        self,
        pbf_file: Optional[Union[str, Path]] = None,
        download_source: PbfSourceLiteral = "protomaps",
        download_directory: Union[str, Path] = "files",
        switch_to_geofabrik_on_error: bool = True,
    ) -> None:
        """
        Initialize OSMPbfLoader.

        Args:
            pbf_file (Union[str, Path], optional): Downloaded `*.osm.pbf` file to be used by
                the loader. If not provided, it will be automatically downloaded for a given area.
                Defaults to None.
            download_source (PbfSourceLiteral, optional): Source to use when downloading PBF files.
                Can be one of: `geofabrik`, `openstreetmap_fr`, `protomaps`.
                Defaults to "protomaps".
            download_directory (Union[str, Path], optional): Directory where to save the downloaded
                `*.osm.pbf` files. Ignored if `pbf_file` is provided. Defaults to "files".
            switch_to_geofabrik_on_error (bool, optional): Flag whether to automatically
                switch `download_source` to 'geofabrik' if error occures. Defaults to `True`.
        """
        import_optional_dependencies(dependency_group="osm", modules=["osmium"])
        self.pbf_file = pbf_file
        self.download_source = download_source
        self.download_directory = download_directory
        self.switch_to_geofabrik_on_error = switch_to_geofabrik_on_error

    def load(
        self,
        area: gpd.GeoDataFrame,
        tags: Union[OsmTagsFilter, GroupedOsmTagsFilter],
    ) -> gpd.GeoDataFrame:
        """
        Load OSM features with specified tags for a given area from an `*.osm.pbf` file.

        The loader will use provided `*.osm.pbf` file, or download extracts
        using `PbfFileDownloader`. Later it will parse and filter features from files
        using `PbfFileHandler`. It will return a GeoDataFrame containing the `geometry` column
        and columns for tag keys.

        Note: Some key/value pairs might be missing from the resulting GeoDataFrame,
            simply because there are no such objects in the given area.

        Args:
            area (gpd.GeoDataFrame): Area for which to download objects.
            tags (Union[OsmTagsFilter, GroupedOsmTagsFilter]): A dictionary
                specifying which tags to download.
                The keys should be OSM tags (e.g. `building`, `amenity`).
                The values should either be `True` for retrieving all objects with the tag,
                string for retrieving a single tag-value pair
                or list of strings for retrieving all values specified in the list.
                `tags={'leisure': 'park}` would return parks from the area.
                `tags={'leisure': 'park, 'amenity': True, 'shop': ['bakery', 'bicycle']}`
                would return parks, all amenity types, bakeries and bicycle shops.

        Raises:
            ValueError: If PBF file is expected to be downloaded and provided geometries
                aren't shapely.geometry.Polygons.

        Returns:
            gpd.GeoDataFrame: Downloaded features as a GeoDataFrame.
        """
        from srai.loaders.osm_loaders.pbf_file_downloader import PbfFileDownloader
        from srai.loaders.osm_loaders.pbf_file_handler import PbfFileHandler

        area_wgs84 = area.to_crs(crs=WGS84_CRS)

        downloaded_pbf_files: Mapping[Hashable, Sequence[Union[str, Path]]]
        if self.pbf_file is not None:
            downloaded_pbf_files = {Path(self.pbf_file).name: [self.pbf_file]}
        else:
            downloaded_pbf_files = PbfFileDownloader(
                download_source=self.download_source,
                download_directory=self.download_directory,
                switch_to_geofabrik_on_error=self.switch_to_geofabrik_on_error,
            ).download_pbf_files_for_regions_gdf(regions_gdf=area_wgs84)

        merged_tags = self._merge_osm_tags_filter(tags)

        pbf_handler = PbfFileHandler(tags=merged_tags)

        results = []
        for region_id, pbf_files in downloaded_pbf_files.items():
            features_gdf = pbf_handler.get_features_gdf(
                file_paths=pbf_files, region_id=str(region_id)
            )
            matching_features_ids = features_gdf.sjoin(area_wgs84).index
            results.append(features_gdf.loc[matching_features_ids])

        result_gdf = self._group_gdfs(results).set_crs(WGS84_CRS)

        features_columns = [
            column
            for column in result_gdf.columns
            if column != GEOMETRY_COLUMN and result_gdf[column].notnull().any()
        ]
        result_gdf = result_gdf[[GEOMETRY_COLUMN, *sorted(features_columns)]]

        return self._parse_features_gdf_to_groups(result_gdf, tags)

    def _group_gdfs(self, gdfs: List[gpd.GeoDataFrame]) -> gpd.GeoDataFrame:
        if not gdfs:
            return self._get_empty_result()
        elif len(gdfs) == 1:
            gdf = gdfs[0]
        else:
            gdf = pd.concat(gdfs)

        return gdf[~gdf.index.duplicated(keep="first")]

    def _get_empty_result(self) -> gpd.GeoDataFrame:
        return gpd.GeoDataFrame(index=pd.Index(name=FEATURES_INDEX), crs=WGS84_CRS, geometry=[])



"""Utility H3 related functions."""

from typing import Iterable, List, Literal, Tuple, Union, overload

import geopandas as gpd
import h3
import numpy as np
import numpy.typing as npt
from h3ronpy.arrow.vector import ContainmentMode, cells_to_wkb_polygons, wkb_to_cells
from shapely.geometry import Polygon
from shapely.geometry.base import BaseGeometry

from srai.constants import GEOMETRY_COLUMN, WGS84_CRS

__all__ = [
    "shapely_geometry_to_h3",
    "h3_to_geoseries",
    "h3_to_shapely_geometry",
    "get_local_ij_index",
]


def shapely_geometry_to_h3(
    geometry: Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame],
    h3_resolution: int,
    buffer: bool = True,
) -> List[str]:
    """
    Convert Shapely geometry to H3 indexes.

    Args:
        geometry (Union[BaseGeometry, Iterable[BaseGeometry], GeoSeries, GeoDataFrame]):
            Shapely geometry to be converted.
        h3_resolution (int): H3 resolution of the cells. See [1] for a full comparison.
        buffer (bool, optional): Whether to fully cover the geometries with
            H3 Cells (visible on the borders). Defaults to True.

    Returns:
        List[str]: List of H3 indexes that cover a given geometry.

    Raises:
        ValueError: If resolution is not between 0 and 15.

    References:
        1. https://h3geo.org/docs/core-library/restable/
    """
    if not (0 <= h3_resolution <= 15):
        raise ValueError(f"Resolution {h3_resolution} is not between 0 and 15.")

    wkb = []
    if isinstance(geometry, gpd.GeoSeries):
        wkb = geometry.to_wkb()
    elif isinstance(geometry, gpd.GeoDataFrame):
        wkb = geometry[GEOMETRY_COLUMN].to_wkb()
    elif isinstance(geometry, Iterable):
        wkb = [sub_geometry.wkb for sub_geometry in geometry]
    else:
        wkb = [geometry.wkb]

    containment_mode = (
        ContainmentMode.IntersectsBoundary if buffer else ContainmentMode.ContainsCentroid
    )
    h3_indexes = wkb_to_cells(
        wkb, resolution=h3_resolution, containment_mode=containment_mode, flatten=True
    ).unique()

    return [h3.int_to_str(h3_index) for h3_index in h3_indexes.tolist()]


# TODO: write tests (#322)
def h3_to_geoseries(h3_index: Union[int, str, Iterable[Union[int, str]]]) -> gpd.GeoSeries:
    """
    Convert H3 index to GeoPandas GeoSeries.

    Args:
        h3_index (Union[int, str, Iterable[Union[int, str]]]): H3 index (or list of indexes)
            to be converted.

    Returns:
        GeoSeries: Geometries as GeoSeries with default CRS applied.
    """
    if isinstance(h3_index, (str, int)):
        return h3_to_geoseries([h3_index])
    else:
        h3_int_indexes = (
            h3_cell if isinstance(h3_cell, int) else h3.str_to_int(h3_cell) for h3_cell in h3_index
        )
        return gpd.GeoSeries.from_wkb(cells_to_wkb_polygons(h3_int_indexes), crs=WGS84_CRS)


@overload
def h3_to_shapely_geometry(h3_index: Union[int, str]) -> Polygon:
    ...


@overload
def h3_to_shapely_geometry(h3_index: Iterable[Union[int, str]]) -> List[Polygon]:
    ...


# TODO: write tests (#322)
def h3_to_shapely_geometry(
    h3_index: Union[int, str, Iterable[Union[int, str]]]
) -> Union[Polygon, List[Polygon]]:
    """
    Convert H3 index to Shapely polygon.

    Args:
        h3_index (Union[int, str, Iterable[Union[int, str]]]): H3 index (or list of indexes)
            to be converted.

    Returns:
        Union[Polygon, List[Polygon]]: Converted polygon (or list of polygons).
    """
    if isinstance(h3_index, (str, int)):
        coords = h3.cell_to_boundary(h3_index, geo_json=True)
        return Polygon(coords)
    return h3_to_geoseries(h3_index).values.tolist()


@overload
def get_local_ij_index(origin_index: str, h3_index: str) -> Tuple[int, int]:
    ...


@overload
def get_local_ij_index(
    origin_index: str, h3_index: List[str], return_as_numpy: Literal[False]
) -> List[Tuple[int, int]]:
    ...


@overload
def get_local_ij_index(
    origin_index: str, h3_index: List[str], return_as_numpy: Literal[True]
) -> npt.NDArray[np.int8]:
    ...


# Last fallback needed as per documentation:
# https://mypy.readthedocs.io/en/stable/literal_types.html#literal-types
@overload
def get_local_ij_index(
    origin_index: str, h3_index: List[str], return_as_numpy: bool
) -> Union[List[Tuple[int, int]], npt.NDArray[np.int8]]:
    ...


def get_local_ij_index(
    origin_index: str, h3_index: Union[str, List[str]], return_as_numpy: bool = False
) -> Union[Tuple[int, int], List[Tuple[int, int]], npt.NDArray[np.int8]]:
    """
    Calculate the local H3 ij index based on provided origin index.

    Wraps H3's cell_to_local_ij function and centers returned coordinates
    around provided origin cell.

    Args:
        origin_index (str): H3 index of the origin region.
        h3_index (Union[str, List[str]]): H3 index of the second region or list of regions.
        return_as_numpy (bool, optional): Flag whether to return calculated indexes as a Numpy array
            or a list of tuples.

    Returns:
        Union[Tuple[int, int], List[Tuple[int, int]], npt.NDArray[np.int8]]: The local ij index of
            the second region (or regions) with respect to the first one.
    """
    origin_coords = h3.cell_to_local_ij(origin_index, origin_index)
    if isinstance(h3_index, str):
        ijs = h3.cell_to_local_ij(origin_index, h3_index)
        return (origin_coords[0] - ijs[0], origin_coords[1] - ijs[1])
    ijs = np.array([h3.cell_to_local_ij(origin_index, h3_cell) for h3_cell in h3_index])
    local_ijs = np.array(origin_coords) - ijs

    if not return_as_numpy:
        local_ijs = [(coords[0], coords[1]) for coords in local_ijs]

    return local_ijs



"""
Folium wrapper.

This module contains functions for quick plotting of analysed gdfs using Geopandas `explore()`
function.
"""
from itertools import cycle, islice
from typing import List, Optional, Set, Union

import branca.colormap as cm
import folium
import geopandas as gpd
import numpy as np
import pandas as pd
import plotly.express as px

from srai._optional import import_optional_dependencies
from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX
from srai.neighbourhoods import Neighbourhood
from srai.neighbourhoods._base import IndexType

import_optional_dependencies(dependency_group="plotting", modules=["folium", "plotly"])


def plot_regions(
    regions_gdf: gpd.GeoDataFrame,
    tiles_style: str = "OpenStreetMap",
    height: Union[str, float] = "100%",
    width: Union[str, float] = "100%",
    colormap: Union[str, List[str]] = px.colors.qualitative.Bold,
    map: Optional[folium.Map] = None,
    show_borders: bool = True,
) -> folium.Map:
    """
    Plot regions shapes using Folium library.

    Args:
        regions_gdf (gpd.GeoDataFrame): Region indexes and geometries to plot.
        tiles_style (str, optional): Map style background. For more styles, look at tiles param at
            https://geopandas.org/en/stable/docs/reference/api/geopandas.GeoDataFrame.explore.html.
            Defaults to "OpenStreetMap".
        height (Union[str, float], optional): Height of the plot. Defaults to "100%".
        width (Union[str, float], optional): Width of the plot. Defaults to "100%".
        colormap (Union[str, List[str]], optional): Colormap to apply to the regions.
            Defaults to `px.colors.qualitative.Bold` from plotly library.
        map (folium.Map, optional): Existing map instance on which to draw the plot.
            Defaults to None.
        show_borders (bool, optional): Whether to show borders between regions or not.
            Defaults to True.

    Returns:
        folium.Map: Generated map.
    """
    return regions_gdf.reset_index().explore(
        column=REGIONS_INDEX,
        tooltip=REGIONS_INDEX,
        tiles=tiles_style,
        height=height,
        width=width,
        legend=False,
        cmap=colormap,
        categorical=True,
        style_kwds=dict(color="#444", opacity=0.5 if show_borders else 0, fillOpacity=0.5),
        m=map,
    )


def plot_numeric_data(
    regions_gdf: gpd.GeoDataFrame,
    data_column: str,
    embedding_df: Optional[Union[pd.DataFrame, gpd.GeoDataFrame]] = None,
    tiles_style: str = "CartoDB positron",
    height: Union[str, float] = "100%",
    width: Union[str, float] = "100%",
    colormap: Union[str, List[str]] = px.colors.sequential.Sunsetdark,
    map: Optional[folium.Map] = None,
    show_borders: bool = False,
    opacity: float = 0.8,
) -> folium.Map:
    """
    Plot numerical data within regions shapes using Folium library.

    Args:
        regions_gdf (gpd.GeoDataFrame): Region indexes and geometries to plot.
        embedding_df (Union[pd.DataFrame, gpd.GeoDataFrame], optional): Region indexes and numerical
            data to plot. If not provided, will use regions_gdf.
        data_column (str): Name of the column used to colour the regions.
        tiles_style (str, optional): Map style background. For more styles, look at tiles param at
            https://geopandas.org/en/stable/docs/reference/api/geopandas.GeoDataFrame.explore.html.
            Defaults to "CartoDB positron".
        height (Union[str, float], optional): Height of the plot. Defaults to "100%".
        width (Union[str, float], optional): Width of the plot. Defaults to "100%".
        colormap (Union[str, List[str]], optional): Colormap to apply to the regions.
            Defaults to px.colors.sequential.Sunsetdark.
        map (folium.Map, optional): Existing map instance on which to draw the plot.
            Defaults to None.
        show_borders (bool, optional): Whether to show borders between regions or not.
            Defaults to False.
        opacity (float, optional): Opacity of coloured regions.

    Returns:
        folium.Map: Generated map.
    """
    if embedding_df is None:
        embedding_df = regions_gdf

    regions_gdf_copy = regions_gdf[[GEOMETRY_COLUMN]].copy()
    regions_gdf_copy = regions_gdf_copy.merge(embedding_df[[data_column]], on=REGIONS_INDEX)

    if not isinstance(colormap, str):
        colormap = _generate_linear_colormap(
            colormap,
            min_value=regions_gdf_copy[data_column].min(),
            max_value=regions_gdf_copy[data_column].max(),
        )

    # Jinja2 rendering issue whne column is number-like. Workaround by using str column name.
    # TypeError: '<' not supported between instances of 'float' and 'str'
    regions_gdf_copy.rename(columns={data_column: "value"}, inplace=True)

    return regions_gdf_copy.reset_index().explore(
        column="value",
        tooltip=[REGIONS_INDEX, "value"],
        tiles=tiles_style,
        height=height,
        width=width,
        legend=True,
        cmap=colormap,
        categorical=False,
        style_kwds=dict(color="#444", opacity=0.5 if show_borders else 0, fillOpacity=opacity),
        m=map,
    )


def plot_neighbours(
    regions_gdf: gpd.GeoDataFrame,
    region_id: IndexType,
    neighbours_ids: Set[IndexType],
    tiles_style: str = "OpenStreetMap",
    height: Union[str, float] = "100%",
    width: Union[str, float] = "100%",
    map: Optional[folium.Map] = None,
    show_borders: bool = True,
) -> folium.Map:
    """
    Plot neighbours on a map using Folium library.

    Args:
        regions_gdf (gpd.GeoDataFrame): Region indexes and geometries to plot.
        region_id (IndexType): Center `region_id` around which the neighbourhood should be plotted.
        neighbours_ids (Set[IndexType]): List of neighbours to highlight.
        tiles_style (str, optional): Map style background. For more styles, look at tiles param at
            https://geopandas.org/en/stable/docs/reference/api/geopandas.GeoDataFrame.explore.html.
            Defaults to "OpenStreetMap".
        height (Union[str, float], optional): Height of the plot. Defaults to "100%".
        width (Union[str, float], optional): Width of the plot. Defaults to "100%".
        map (folium.Map, optional): Existing map instance on which to draw the plot.
            Defaults to None.
        show_borders (bool, optional): Whether to show borders between regions or not.
            Defaults to True.

    Returns:
        folium.Map: Generated map.
    """
    if region_id not in regions_gdf.index:
        raise ValueError(f"{region_id!r} doesn't exist in provided regions_gdf.")

    regions_gdf_copy = regions_gdf.copy()
    regions_gdf_copy["region"] = "other"
    regions_gdf_copy.loc[region_id, "region"] = "selected"
    regions_gdf_copy.loc[list(neighbours_ids), "region"] = "neighbour"
    return regions_gdf_copy.reset_index().explore(
        column="region",
        tooltip=REGIONS_INDEX,
        tiles=tiles_style,
        height=height,
        width=width,
        cmap=[
            "rgb(242, 242, 242)",
            px.colors.sequential.Sunsetdark[-1],
            px.colors.sequential.Sunsetdark[2],
        ],
        categorical=True,
        categories=["selected", "neighbour", "other"],
        style_kwds=dict(color="#444", opacity=0.5 if show_borders else 0, fillOpacity=0.8),
        m=map,
    )


def plot_all_neighbourhood(
    regions_gdf: gpd.GeoDataFrame,
    region_id: IndexType,
    neighbourhood: Neighbourhood[IndexType],
    neighbourhood_max_distance: int = 100,
    tiles_style: str = "OpenStreetMap",
    height: Union[str, float] = "100%",
    width: Union[str, float] = "100%",
    colormap: Union[str, List[str]] = px.colors.sequential.Agsunset_r,
    map: Optional[folium.Map] = None,
    show_borders: bool = True,
) -> folium.Map:
    """
    Plot full neighbourhood on a map using Folium library.

    Args:
        regions_gdf (gpd.GeoDataFrame): Region indexes and geometries to plot.
        region_id (IndexType): Center `region_id` around which the neighbourhood should be plotted.
        neighbourhood (Neighbourhood[IndexType]): `Neighbourhood` class required for finding
            neighbours.
        neighbourhood_max_distance (int, optional): Max distance for rendering neighbourhoods.
            Neighbours farther away won't be coloured, and will be left as "other" regions.
            Defaults to 100.
        tiles_style (str, optional): Map style background. For more styles, look at tiles param at
            https://geopandas.org/en/stable/docs/reference/api/geopandas.GeoDataFrame.explore.html.
            Defaults to "OpenStreetMap".
        height (Union[str, float], optional): Height of the plot. Defaults to "100%".
        width (Union[str, float], optional): Width of the plot. Defaults to "100%".
        colormap (Union[str, List[str]], optional): Colormap to apply to the neighbourhoods.
            Defaults to `px.colors.sequential.Agsunset_r` from plotly library.
        map (folium.Map, optional): Existing map instance on which to draw the plot.
            Defaults to None.
        show_borders (bool, optional): Whether to show borders between regions or not.
            Defaults to True.

    Returns:
        folium.Map: Generated map.
    """
    if region_id not in regions_gdf.index:
        raise ValueError(f"{region_id!r} doesn't exist in provided regions_gdf.")

    regions_gdf_copy = regions_gdf.copy()
    regions_gdf_copy["region"] = "other"
    regions_gdf_copy.loc[region_id, "region"] = "selected"

    distance = 1
    neighbours_ids = neighbourhood.get_neighbours_at_distance(region_id, distance).intersection(
        regions_gdf.index
    )
    while neighbours_ids and distance <= neighbourhood_max_distance:
        regions_gdf_copy.loc[list(neighbours_ids), "region"] = distance
        distance += 1
        neighbours_ids = neighbourhood.get_neighbours_at_distance(region_id, distance).intersection(
            regions_gdf.index
        )

    if not isinstance(colormap, str):
        colormap = _generate_colormap(
            distance, colormap=_resample_plotly_colormap(colormap, min(distance, 10))
        )

    return regions_gdf_copy.reset_index().explore(
        column="region",
        tooltip=[REGIONS_INDEX, "region"],
        tiles=tiles_style,
        height=height,
        width=width,
        cmap=colormap,
        categorical=True,
        categories=["selected", *list(range(distance))[1:], "other"],
        style_kwds=dict(color="#444", opacity=0.5 if show_borders else 0, fillOpacity=0.8),
        legend=distance <= 11,
        m=map,
    )


def _resample_plotly_colormap(colormap: List[str], steps: int) -> List[str]:
    resampled_colormap: List[str] = px.colors.sample_colorscale(
        colormap, np.linspace(0, 1, num=steps)
    )
    return resampled_colormap


def _generate_colormap(
    distance: int,
    colormap: List[str],
    selected_color: str = "rgb(242, 242, 242)",
    other_color: str = "rgb(153, 153, 153)",
) -> List[str]:
    return [selected_color, *islice(cycle(colormap), None, distance - 1), other_color]


def _generate_linear_colormap(
    colormap: List[str], min_value: float, max_value: float
) -> cm.LinearColormap:
    values, _ = px.colors.convert_colors_to_same_type(colormap, colortype="tuple")
    return cm.LinearColormap(values, vmin=min_value, vmax=max_value)



"""
PBF File Downloader.

This module contains a downloader capable of downloading a PBF files from multiple sources.
"""
import json
import time
import warnings
from pathlib import Path
from time import sleep
from typing import Any, Callable, Dict, Hashable, List, Literal, Sequence, Union

import geopandas as gpd
import requests
import topojson as tp
from requests import HTTPError
from shapely.geometry import Polygon, mapping
from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry
from tqdm import tqdm

from srai.constants import WGS84_CRS
from srai.geometry import (
    buffer_geometry,
    flatten_geometry,
    flatten_geometry_series,
    get_geometry_hash,
    remove_interiors,
)
from srai.loaders import download_file
from srai.loaders.osm_loaders.openstreetmap_extracts import (
    OpenStreetMapExtract,
    find_smallest_containing_geofabrik_extracts,
    find_smallest_containing_openstreetmap_fr_extracts,
)
from srai.loaders.osm_loaders.pbf_file_clipper import PbfFileClipper

PbfSourceLiteral = Literal["geofabrik", "openstreetmap_fr", "protomaps"]
PbfSourceExtractsFunctions: Dict[
    PbfSourceLiteral,
    Callable[[Union[BaseGeometry, BaseMultipartGeometry]], List[OpenStreetMapExtract]],
] = {
    "geofabrik": find_smallest_containing_geofabrik_extracts,
    "openstreetmap_fr": find_smallest_containing_openstreetmap_fr_extracts,
}


class PbfFileDownloader:
    """
    PbfFileDownloader.

    PBF(Protocolbuffer Binary Format)[1] file downloader is a downloader
    capable of downloading `*.osm.pbf` files with OSM data for a given area.

    This downloader can use multiple sources to extract a PBF file for a given region:
     - Geofabrik - free hosting service with PBF files http://download.geofabrik.de/.
     - OpenStreetMap.fr - free hosting service with PBF files https://download.openstreetmap.fr/.
     - Protomaps - (will be deprecated!) free download service for downloading an extract for
       an area of interest.


    References:
        1. https://wiki.openstreetmap.org/wiki/PBF_Format
    """

    PROTOMAPS_API_START_URL = "https://app.protomaps.com/downloads/osm"
    PROTOMAPS_API_DOWNLOAD_URL = "https://app.protomaps.com/downloads/{}/download"
    PROTOMAPS_MAX_WAIT_TIME_S = 300  # max 5 minutes for protomaps to generate an extract

    _PBAR_FORMAT = "[{}] Downloading pbf file #{} ({})"

    SIMPLIFICATION_TOLERANCE_VALUES = [
        1e-07,
        2e-07,
        5e-07,
        1e-06,
        2e-06,
        5e-06,
        1e-05,
        2e-05,
        5e-05,
        0.0001,
        0.0002,
        0.0005,
        0.001,
        0.002,
        0.005,
        0.01,
        0.02,
        0.05,
    ]

    def __init__(
        self,
        download_source: PbfSourceLiteral = "protomaps",
        download_directory: Union[str, Path] = "files",
        switch_to_geofabrik_on_error: bool = True,
    ) -> None:
        """
        Initialize PbfFileDownloader.

        Args:
            download_source (PbfSourceLiteral, optional): Source to use when downloading PBF files.
                Can be one of: `geofabrik`, `openstreetmap_fr`, `protomaps`.
                Defaults to "protomaps".
            download_directory (Union[str, Path], optional): Directory where to save
                the downloaded `*.osm.pbf` files. Defaults to "files".
            switch_to_geofabrik_on_error (bool, optional): Flag whether to automatically
                switch `download_source` to 'geofabrik' if error occures. Defaults to `True`.
        """
        self.download_source = download_source
        self.download_directory = download_directory
        self.clipper = PbfFileClipper(working_directory=self.download_directory)
        self.switch_to_geofabrik_on_error = switch_to_geofabrik_on_error

    def download_pbf_files_for_regions_gdf(
        self, regions_gdf: gpd.GeoDataFrame
    ) -> Dict[Hashable, Sequence[Path]]:
        """
        Download PBF files for regions GeoDataFrame.

        Function will split each multipolygon into single polygons and download PBF files
        for each of them.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.

        Raises:
            ValueError: If provided geometries aren't shapely.geometry.Polygons.

        Returns:
            Dict[Hashable, Sequence[Path]]: List of Paths to downloaded PBF files per
                each region_id.
        """
        regions_mapping: Dict[Hashable, Sequence[Path]] = {}

        non_polygon_types = set(
            type(geometry)
            for geometry in flatten_geometry_series(regions_gdf.geometry)
            if not isinstance(geometry, Polygon)
        )
        if non_polygon_types:
            raise ValueError(f"Provided geometries aren't Polygons (found: {non_polygon_types})")

        try:
            if self.download_source == "protomaps":
                regions_mapping = self._download_pbf_files_for_polygons_from_protomaps(regions_gdf)
            elif self.download_source in PbfSourceExtractsFunctions:
                regions_mapping = self._download_pbf_files_for_polygons_from_existing_extracts(
                    regions_gdf
                )
        except Exception as err:
            if self.download_source != "geofabrik" and self.switch_to_geofabrik_on_error:
                warnings.warn(
                    f"Error occured ({err}). Auto-switching to 'geofabrik' download source.",
                    stacklevel=1,
                )
                regions_mapping = self._download_pbf_files_for_polygons_from_existing_extracts(
                    regions_gdf, override_to_geofabrik=True
                )
            else:
                error_message = str(err)
                if self.download_source != "geofabrik":
                    error_message += (
                        "\nPlease change the 'download_source' to"
                        " 'geofabrik' or other availablesource:\n"
                        " PbfDownloader(download_source='geofabrik', ...) or"
                        " OsmPbfLoader(download_source='geofabrik', ...)."
                    )
                raise RuntimeError(error_message) from err

        return regions_mapping

    def _download_pbf_files_for_polygons_from_existing_extracts(
        self, regions_gdf: gpd.GeoDataFrame, override_to_geofabrik: bool = False
    ) -> Dict[Hashable, Sequence[Path]]:
        regions_mapping: Dict[Hashable, Sequence[Path]] = {}

        unary_union_geometry = regions_gdf.geometry.unary_union

        if override_to_geofabrik:
            extract_function = PbfSourceExtractsFunctions["geofabrik"]
        else:
            extract_function = PbfSourceExtractsFunctions[self.download_source]
        extracts = extract_function(unary_union_geometry)

        downloaded_pbf_files = []

        for extract in extracts:
            pbf_file_path = Path(self.download_directory).resolve() / f"{extract.id}.osm.pbf"

            download_file(url=extract.url, fname=pbf_file_path.as_posix(), force_download=False)

            downloaded_pbf_files.append(pbf_file_path)

        polygons = flatten_geometry(unary_union_geometry)

        for region_id, row in regions_gdf.iterrows():
            polygons = flatten_geometry(row.geometry)
            regions_mapping[region_id] = [
                self.clipper.clip_pbf_file(polygon, downloaded_pbf_files) for polygon in polygons
            ]

        return regions_mapping

    def _download_pbf_files_for_polygons_from_protomaps(
        self, regions_gdf: gpd.GeoDataFrame
    ) -> Dict[Hashable, Sequence[Path]]:
        regions_mapping: Dict[Hashable, Sequence[Path]] = {}

        for region_id, row in regions_gdf.iterrows():
            polygons = flatten_geometry(row.geometry)
            regions_mapping[region_id] = [
                self._download_pbf_file_for_polygon_from_protomaps(
                    polygon, region_id, polygon_id + 1
                )
                for polygon_id, polygon in enumerate(polygons)
            ]

        return regions_mapping

    def _download_pbf_file_for_polygon_from_protomaps(
        self, polygon: Polygon, region_id: str = "OSM", polygon_id: int = 1
    ) -> Path:
        """
        Download PBF file for a single Polygon.

        Function will buffer polygon by 50 meters, simplify exterior boundary to be
        below 1000 points (which is a limit of Protomaps API) and close all holes within it.

        Boundary of the polygon will be sent to Protomaps service and an `*.osm.pbf` file
        will be downloaded with a hash based on WKT representation of the parsed polygon.
        If file exists, it won't be downloaded again.

        Args:
            polygon (Polygon): Polygon boundary of an area to be extracted.
            region_id (str, optional): Region name to be set in progress bar.
                Defaults to "OSM".
            polygon_id (int, optional): Polygon number to be set in progress bar.
                Defaults to 1.

        Returns:
            Path: Path to a downloaded `*.osm.pbf` file.
        """
        geometry_hash = get_geometry_hash(polygon)
        pbf_file_path = Path(self.download_directory).resolve() / f"{geometry_hash}.osm.pbf"

        if not pbf_file_path.exists():  # pragma: no cover
            boundary_polygon = self._prepare_polygon_for_download(polygon)
            geometry_geojson = mapping(boundary_polygon)

            s = requests.Session()

            req = s.get(url=self.PROTOMAPS_API_START_URL)

            csrf_token = req.cookies["csrftoken"]
            headers = {
                "Referer": self.PROTOMAPS_API_START_URL,
                "Cookie": f"csrftoken={csrf_token}",
                "X-CSRFToken": csrf_token,
                "Content-Type": "application/json; charset=utf-8",
                "User-Agent": "SRAI Python package (https://github.com/kraina-ai/srai)",
            }
            request_payload = {
                "region": {"type": "geojson", "data": geometry_geojson},
                "name": geometry_hash,
            }

            start_extract_request = s.post(
                url=self.PROTOMAPS_API_START_URL,
                json=request_payload,
                headers=headers,
                cookies=dict(csrftoken=csrf_token),
            )
            start_extract_request.raise_for_status()

            start_extract_result = start_extract_request.json()
            try:
                extraction_uuid = start_extract_result["uuid"]
                status_check_url = start_extract_result["url"]
            except KeyError as err:
                error_message = (
                    f"Error from the 'Protomaps' service: {json.dumps(start_extract_result)}."
                )
                raise RuntimeError(error_message) from err

            with tqdm() as pbar:
                status_response: Dict[str, Any] = {}
                cells_total = 0
                nodes_total = 0
                elems_total = 0
                start_time = time.time()
                while not status_response.get("Complete", False):
                    sleep(0.5)
                    status_response = s.get(url=status_check_url).json()
                    cells_total = max(cells_total, status_response.get("CellsTotal", 0))
                    nodes_total = max(nodes_total, status_response.get("NodesTotal", 0))
                    elems_total = max(elems_total, status_response.get("ElemsTotal", 0))

                    cells_prog = status_response.get("CellsProg", None)
                    nodes_prog = status_response.get("NodesProg", None)
                    elems_prog = status_response.get("ElemsProg", None)

                    if cells_total > 0 and cells_prog is not None and cells_prog < cells_total:
                        pbar.set_description(
                            self._PBAR_FORMAT.format(region_id, polygon_id, "Cells")
                        )
                        pbar.total = cells_total + nodes_total + elems_total
                        pbar.n = cells_prog
                    elif nodes_total > 0 and nodes_prog is not None and nodes_prog < nodes_total:
                        pbar.set_description(
                            self._PBAR_FORMAT.format(region_id, polygon_id, "Nodes")
                        )
                        pbar.total = cells_total + nodes_total + elems_total
                        pbar.n = cells_total + nodes_prog
                    elif elems_total > 0 and elems_prog is not None and elems_prog < elems_total:
                        pbar.set_description(
                            self._PBAR_FORMAT.format(region_id, polygon_id, "Elements")
                        )
                        pbar.total = cells_total + nodes_total + elems_total
                        pbar.n = cells_total + nodes_total + elems_prog
                    else:
                        pbar.total = cells_total + nodes_total + elems_total
                        pbar.n = cells_total + nodes_total + elems_total

                    pbar.refresh()

                    if (time.time() - start_time) > self.PROTOMAPS_MAX_WAIT_TIME_S:
                        error_message = (
                            "'Protomaps' service took too long to generate an extract"
                            f" ({self.PROTOMAPS_MAX_WAIT_TIME_S}s)."
                        )
                        raise RuntimeError(error_message)

            try:
                download_file(
                    url=self.PROTOMAPS_API_DOWNLOAD_URL.format(extraction_uuid),
                    fname=pbf_file_path.as_posix(),
                )
            except HTTPError as err:
                error_message = f"Error from the 'Protomaps' service: {err.response}."
                raise RuntimeError(error_message) from err

        return pbf_file_path

    def _prepare_polygon_for_download(self, polygon: Polygon) -> Polygon:
        """
        Prepare polygon for download.

        Function buffers the polygon, closes internal holes and simplifies its boundary to 1000
        points.

        Makes sure that the generated polygon with fully cover the original one by increasing the
        buffer size incrementally. Buffering is applied to the last simplified geometry to speed up
        the process.
        """
        is_fully_covered = False
        buffer_size_meters = 50

        polygon_to_buffer = polygon

        while not is_fully_covered:
            buffered_polygon = buffer_geometry(polygon_to_buffer, meters=buffer_size_meters)
            simplified_polygon = self._simplify_polygon(buffered_polygon, 1000)
            closed_polygon = remove_interiors(simplified_polygon)
            is_fully_covered = polygon.covered_by(closed_polygon)
            buffer_size_meters += 50

            polygon_to_buffer = closed_polygon

        return closed_polygon

    def _simplify_polygon(self, polygon: Polygon, exterior_max_points: int = 1000) -> Polygon:
        """Simplify a polygon boundary to up to provided number of points."""
        simplified_polygon = polygon

        for simplify_tolerance in self.SIMPLIFICATION_TOLERANCE_VALUES:
            simplified_polygon = (
                tp.Topology(
                    polygon,
                    toposimplify=simplify_tolerance,
                    prevent_oversimplify=True,
                )
                .to_gdf(winding_order="CW_CCW", crs=WGS84_CRS, validate=True)
                .geometry[0]
            )

            if len(simplified_polygon.exterior.coords) < exterior_max_points:
                break

        if len(simplified_polygon.exterior.coords) > exterior_max_points:
            simplified_polygon = polygon.convex_hull

        if len(simplified_polygon.exterior.coords) > exterior_max_points:
            simplified_polygon = polygon.minimum_rotated_rectangle

        return simplified_polygon



"""
Geoparquet loader.

This module contains geoparquet loader implementation.
"""

from pathlib import Path
from typing import List, Optional, Union

import geopandas as gpd

from srai.constants import GEOMETRY_COLUMN, WGS84_CRS
from srai.loaders import Loader


class GeoparquetLoader(Loader):
    """
    GeoparquetLoader.

    Geoparquet [1] loader is a wrapper for a `geopandas.read_parquet` function
    and allows for an automatic index setting and additional geometry clipping.

    References:
        1. https://github.com/opengeospatial/geoparquet
    """

    def load(
        self,
        file_path: Union[Path, str],
        index_column: Optional[str] = None,
        columns: Optional[List[str]] = None,
        area: Optional[gpd.GeoDataFrame] = None,
    ) -> gpd.GeoDataFrame:
        """
        Load a geoparquet file.

        Args:
            file_path (Union[Path, str]): parquet file path.
            index_column (str, optional): Column that will be used as an index.
                If not provided, automatic indexing will be applied by default. Defaults to None.
            columns (List[str], optional): List of columns to load.
                If not provided, all will be loaded. Defaults to None.
            area (gpd.GeoDataFrame, optional): Mask to clip loaded data.
                If not provided, unaltered data will be returned. Defaults to None.

        Raises:
            ValueError: If provided index column doesn't exists in list of loaded columns.

        Returns:
            gpd.GeoDataFrame: Loaded geoparquet file as a GeoDataFrame.
        """
        if columns and GEOMETRY_COLUMN not in columns:
            columns.append(GEOMETRY_COLUMN)

        gdf = gpd.read_parquet(path=file_path, columns=columns)

        if index_column:
            if index_column not in gdf.columns:
                raise ValueError(f"Column {index_column} doesn't exist in a file.")
            gdf.set_index(index_column, inplace=True)

        gdf.to_crs(crs=WGS84_CRS, inplace=True)

        if area is not None:
            area_wgs84 = area.to_crs(crs=WGS84_CRS)
            gdf = gdf.clip(mask=area_wgs84, keep_geom_type=False)

        return gdf



"""Utility geometry operations functions."""
import hashlib
from typing import List, Union

import geopandas as gpd
import pyproj
import shapely.wkt as wktlib
from functional import seq
from shapely.geometry import MultiPolygon, Polygon
from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry
from shapely.ops import transform as shapely_transform

__all__ = [
    "flatten_geometry_series",
    "flatten_geometry",
    "remove_interiors",
    "buffer_geometry",
    "merge_disjointed_polygons",
    "merge_disjointed_gdf_geometries",
]


def flatten_geometry_series(geometry_series: gpd.GeoSeries) -> List[BaseGeometry]:
    """Flatten all geometries from a series into a list of BaseGeometries."""
    geometries: List[BaseGeometry] = (
        seq([flatten_geometry(geometry) for geometry in geometry_series]).flatten().to_list()
    )
    return geometries


def flatten_geometry(geometry: BaseGeometry) -> List[BaseGeometry]:
    """Flatten all geometries into a list of BaseGeometries."""
    if isinstance(geometry, BaseMultipartGeometry):
        geometries: List[BaseGeometry] = (
            seq([flatten_geometry(sub_geom) for sub_geom in geometry.geoms]).flatten().to_list()
        )
        return geometries
    return [geometry]


# https://stackoverflow.com/a/70387141/7766101
def remove_interiors(polygon: Polygon) -> Polygon:
    """
    Close polygon holes by limitation to the exterior ring.

    Args:
        polygon (Polygon): Polygon to close.

    Returns:
        Polygon: Closed polygon.
    """
    if polygon.interiors:
        return Polygon(list(polygon.exterior.coords))
    return polygon


def buffer_geometry(geometry: BaseGeometry, meters: float) -> BaseGeometry:
    """
    Buffer geometry by a given radius in meters.

    Projects geometry into azimuthal projection before applying buffer and then changes values
    back to WGS84 coordinates.

    Doesn't work with polygons covering the whole earth (from -180 to 180 longitude).

    Args:
        geometry (BaseGeometry): Geometry to buffer.
        meters (float): Radius distance in meters.

    Returns:
        BaseGeometry: Buffered geometry.
    """
    _lon, _lat = geometry.centroid.coords[0]

    aeqd_proj = pyproj.Proj(proj="aeqd", ellps="WGS84", datum="WGS84", lat_0=_lat, lon_0=_lon)
    wgs84_proj = pyproj.Proj(proj="latlong", ellps="WGS84")

    wgs84_to_aeqd = pyproj.Transformer.from_proj(wgs84_proj, aeqd_proj, always_xy=True).transform
    aeqd_to_wgs84 = pyproj.Transformer.from_proj(aeqd_proj, wgs84_proj, always_xy=True).transform

    projected_geometry = shapely_transform(wgs84_to_aeqd, geometry)
    bufferred_projected_geometry = projected_geometry.buffer(meters)

    return shapely_transform(aeqd_to_wgs84, bufferred_projected_geometry)


def merge_disjointed_polygons(polygons: List[Union[Polygon, MultiPolygon]]) -> MultiPolygon:
    """
    Merges all polygons into a single MultiPolygon.

    Input polygons are expected to be disjointed.

    Args:
        polygons (List[Union[Polygon, MultiPolygon]]): List of polygons to merge

    Returns:
        MultiPolygon: Merged polygon
    """
    single_polygons = []
    for geom in polygons:
        if isinstance(geom, Polygon):
            single_polygons.append(geom)
        else:
            single_polygons.extend(geom.geoms)
    return MultiPolygon(single_polygons)


def merge_disjointed_gdf_geometries(gdf: gpd.GeoDataFrame) -> MultiPolygon:
    """
    Merges geometries from a GeoDataFrame into a single MultiPolygon.

    Input geometries are expected to be disjointed.

    Args:
        gdf (gpd.GeoDataFrame): GeoDataFrame with geometries to merge.

    Returns:
        MultiPolygon: Merged polygon
    """
    return merge_disjointed_polygons(list(gdf.geometry))


def get_geometry_hash(geometry: BaseGeometry) -> str:
    """Generate SHA256 hash based on WKT representation of the polygon."""
    wkt_string = wktlib.dumps(geometry)
    h = hashlib.new("sha256")
    h.update(wkt_string.encode())
    return h.hexdigest()



"""
Plotly wrapper.

This module contains functions for quick plotting of analysed gdfs using Plotly library.
"""
from typing import Any, Dict, List, Optional, Set

import geopandas as gpd
import numpy as np
import plotly.express as px
import plotly.graph_objs as go
from shapely.geometry import Point

from srai._optional import import_optional_dependencies
from srai.constants import REGIONS_INDEX, WGS84_CRS
from srai.neighbourhoods import Neighbourhood
from srai.neighbourhoods._base import IndexType

import_optional_dependencies(dependency_group="plotting", modules=["plotly"])


def plot_regions(
    regions_gdf: gpd.GeoDataFrame,
    return_plot: bool = False,
    mapbox_style: str = "open-street-map",
    mapbox_accesstoken: Optional[str] = None,
    renderer: Optional[str] = None,
    zoom: Optional[float] = None,
    height: Optional[float] = None,
    width: Optional[float] = None,
) -> Optional[go.Figure]:
    """
    Plot regions shapes using Plotly library.

    For more info about parameters, check https://plotly.com/python/mapbox-layers/.

    Args:
        regions_gdf (gpd.GeoDataFrame): Region indexes and geometries to plot.
        return_plot (bool, optional): Flag whether to return the Figure object or not.
            If `True`, the plot won't be displayed automatically. Defaults to False.
        mapbox_style (str, optional): Map style background. Defaults to "open-street-map".
        mapbox_accesstoken (str, optional): Access token required for mapbox based map backgrounds.
            Defaults to None.
        renderer (str, optional): Name of renderer used for displaying the figure.
            For all descriptions, look here: https://plotly.com/python/renderers/.
            Defaults to None.
        zoom (float, optional): Map zoom. If not filled, will be approximated based on
            the bounding box of regions. Defaults to None.
        height (float, optional): Height of the plot. Defaults to None.
        width (float, optional): Width of the plot. Defaults to None.

    Returns:
        Optional[go.Figure]: Figure of the plot. Will be returned if `return_plot` is set to `True`.
    """
    regions_gdf_copy = regions_gdf.copy()
    regions_gdf_copy[REGIONS_INDEX] = regions_gdf_copy.index
    return _plot_regions(
        regions_gdf=regions_gdf_copy,
        hover_column_name=REGIONS_INDEX,
        color_feature_column=None,
        hover_data=[],
        show_legend=False,
        return_plot=return_plot,
        mapbox_style=mapbox_style,
        mapbox_accesstoken=mapbox_accesstoken,
        renderer=renderer,
        zoom=zoom,
        height=height,
        width=width,
        color_discrete_sequence=px.colors.qualitative.Safe,
        opacity=0.4,
        traces_kwargs=dict(marker_line_width=2),
    )


def plot_neighbours(
    regions_gdf: gpd.GeoDataFrame,
    region_id: IndexType,
    neighbours_ids: Set[IndexType],
    return_plot: bool = False,
    mapbox_style: str = "open-street-map",
    mapbox_accesstoken: Optional[str] = None,
    renderer: Optional[str] = None,
    zoom: Optional[float] = None,
    height: Optional[float] = None,
    width: Optional[float] = None,
) -> Optional[go.Figure]:
    """
    Plot neighbours on a map using Plotly library.

    For more info about parameters, check https://plotly.com/python/mapbox-layers/.

    Args:
        regions_gdf (gpd.GeoDataFrame): Region indexes and geometries to plot.
        region_id (IndexType): Center `region_id` around which the neighbourhood should be plotted.
        neighbours_ids (Set[IndexType]): List of neighbours to highlight.
        return_plot (bool, optional): Flag whether to return the Figure object or not.
            If `True`, the plot won't be displayed automatically. Defaults to False.
        mapbox_style (str, optional): Map style background. Defaults to "open-street-map".
        mapbox_accesstoken (str, optional): Access token required for mapbox based map backgrounds.
            Defaults to None.
        renderer (str, optional): Name of renderer used for displaying the figure.
            For all descriptions, look here: https://plotly.com/python/renderers/.
            Defaults to None.
        zoom (float, optional): Map zoom. If not filled, will be approximated based on
            the bounding box of regions. Defaults to None.
        height (float, optional): Height of the plot. Defaults to None.
        width (float, optional): Width of the plot. Defaults to None.

    Returns:
        Optional[go.Figure]: Figure of the plot. Will be returned if `return_plot` is set to `True`.
    """
    regions_gdf_copy = regions_gdf.copy()
    regions_gdf_copy[REGIONS_INDEX] = regions_gdf_copy.index
    regions_gdf_copy["region"] = "other"
    regions_gdf_copy.loc[region_id, "region"] = "selected"
    regions_gdf_copy.loc[list(neighbours_ids), "region"] = "neighbour"
    return _plot_regions(
        regions_gdf=regions_gdf_copy,
        hover_column_name=REGIONS_INDEX,
        color_feature_column="region",
        hover_data=[],
        show_legend=True,
        return_plot=return_plot,
        mapbox_style=mapbox_style,
        mapbox_accesstoken=mapbox_accesstoken,
        renderer=renderer,
        zoom=zoom,
        height=height,
        width=width,
        layout_kwargs=dict(
            legend=dict(yanchor="top", y=0.99, xanchor="left", x=0.01, traceorder="normal"),
        ),
        category_orders={"region": ["selected", "neighbour", "other"]},
        color_discrete_sequence=[
            px.colors.qualitative.Plotly[1],
            px.colors.qualitative.Plotly[2],
            px.colors.qualitative.Plotly[0],
        ],
    )


def plot_all_neighbourhood(
    regions_gdf: gpd.GeoDataFrame,
    region_id: IndexType,
    neighbourhood: Neighbourhood[IndexType],
    neighbourhood_max_distance: int = 100,
    return_plot: bool = False,
    mapbox_style: str = "open-street-map",
    mapbox_accesstoken: Optional[str] = None,
    renderer: Optional[str] = None,
    zoom: Optional[float] = None,
    height: Optional[float] = None,
    width: Optional[float] = None,
) -> Optional[go.Figure]:
    """
    Plot full neighbourhood on a map using Plotly library.

    For more info about parameters, check https://plotly.com/python/mapbox-layers/.

    Args:
        regions_gdf (gpd.GeoDataFrame): Region indexes and geometries to plot.
        region_id (IndexType): Center `region_id` around which the neighbourhood should be plotted.
        neighbourhood (Neighbourhood[IndexType]): `Neighbourhood` class required for finding
            neighbours.
        neighbourhood_max_distance (int, optional): Max distance for rendering neighbourhoods.
            Neighbours farther away won't be coloured, and will be left as "other" regions.
            Defaults to 100.
        return_plot (bool, optional): Flag whether to return the Figure object or not.
            If `True`, the plot won't be displayed automatically. Defaults to False.
        mapbox_style (str, optional): Map style background. Defaults to "open-street-map".
        mapbox_accesstoken (str, optional): Access token required for mapbox based map backgrounds.
            Defaults to None.
        renderer (str, optional): Name of renderer used for displaying the figure.
            For all descriptions, look here: https://plotly.com/python/renderers/.
            Defaults to None.
        zoom (float, optional): Map zoom. If not filled, will be approximated based on
            the bounding box of regions. Defaults to None.
        height (float, optional): Height of the plot. Defaults to None.
        width (float, optional): Width of the plot. Defaults to None.

    Returns:
        Optional[go.Figure]: Figure of the plot. Will be returned if `return_plot` is set to `True`.
    """
    regions_gdf_copy = regions_gdf.copy()
    regions_gdf_copy[REGIONS_INDEX] = regions_gdf_copy.index
    regions_gdf_copy["region"] = "other"
    regions_gdf_copy.loc[region_id, "region"] = "selected"

    distance = 1
    neighbours_ids = neighbourhood.get_neighbours_at_distance(region_id, distance).intersection(
        regions_gdf.index
    )
    while neighbours_ids and distance <= neighbourhood_max_distance:
        regions_gdf_copy.loc[list(neighbours_ids), "region"] = distance
        distance += 1
        neighbours_ids = neighbourhood.get_neighbours_at_distance(region_id, distance).intersection(
            regions_gdf.index
        )

    return _plot_regions(
        regions_gdf=regions_gdf_copy,
        hover_column_name=REGIONS_INDEX,
        color_feature_column="region",
        hover_data=[],
        show_legend=True,
        return_plot=return_plot,
        mapbox_style=mapbox_style,
        mapbox_accesstoken=mapbox_accesstoken,
        renderer=renderer,
        zoom=zoom,
        height=height,
        width=width,
        layout_kwargs=dict(
            legend=dict(yanchor="top", y=0.99, xanchor="left", x=0.01, traceorder="normal"),
        ),
        category_orders={"region": ["selected", *range(distance), "other"]},
        color_discrete_sequence=px.colors.cyclical.Edge,
    )


def _plot_regions(
    regions_gdf: gpd.GeoDataFrame,
    hover_column_name: str,
    hover_data: List[str],
    color_feature_column: Optional[str] = None,
    show_legend: bool = False,
    return_plot: bool = False,
    mapbox_style: str = "open-street-map",
    mapbox_accesstoken: Optional[str] = None,
    opacity: float = 0.6,
    renderer: Optional[str] = None,
    zoom: Optional[float] = None,
    height: Optional[float] = None,
    width: Optional[float] = None,
    layout_kwargs: Optional[Dict[str, Any]] = None,
    traces_kwargs: Optional[Dict[str, Any]] = None,
    **choropleth_mapbox_kwargs: Any,
) -> Optional[go.Figure]:
    """
    Plot regions shapes using Plotly library.

    Uses `choroplethmapbox` function.
    For more info about parameters, check https://plotly.com/python/mapbox-layers/.

    Args:
        regions_gdf (gpd.GeoDataFrame): Region indexes and geometries to plot.
        hover_column_name (str): Column name used for hover popup title.
        hover_data (List[str]): List of column names displayed additionally on hover.
        color_feature_column (str, optional): Column name used for colouring the plot.
            Can be `None` to disable colouring.
        show_legend (bool, optional): Flag whether to show the legend or not. Defaults to False.
        return_plot (bool, optional): Flag whether to return the Figure object or not.
            If `True`, the plot won't be displayed automatically. Defaults to False.
        mapbox_style (str, optional): Map style background. Defaults to "open-street-map".
        mapbox_accesstoken (str, optional): Access token required for mapbox based map backgrounds.
            Defaults to None.
        opacity (float, optional): Markers opacity. Defaults to 0.6.
        renderer (str, optional): Name of renderer used for displaying the figure.
            For all descriptions, look here: https://plotly.com/python/renderers/.
            Defaults to None.
        zoom (float, optional): Map zoom. If not filled, will be approximated based on
            the bounding box of regions. Defaults to None.
        height (float, optional): Height of the plot. Defaults to None.
        width (float, optional): Width of the plot. Defaults to None.
        layout_kwargs (Dict[str, Any], optional): Additional parameters passed to
            the `update_layout` function. Defaults to None.
        traces_kwargs (Dict[str, Any], optional): Additional parameters passed to
            the `update_traces` function. Defaults to None.
        **choropleth_mapbox_kwargs: Additional parameters that can be passed to
            the `choropleth_mapbox` constructor.

    Returns:
        Optional[go.Figure]: Figure of the plot. Will be returned if `return_plot` is set to `True`.
    """
    center_point = _calculate_map_centroid(regions_gdf)
    if not zoom:
        zoom = _calculate_mapbox_zoom(regions_gdf)

    fig = px.choropleth_mapbox(
        regions_gdf,
        geojson=regions_gdf,
        color=color_feature_column,
        hover_name=hover_column_name,
        hover_data=hover_data,
        locations=REGIONS_INDEX,
        center={"lon": center_point.x, "lat": center_point.y},
        zoom=zoom,
        **choropleth_mapbox_kwargs,
    )

    update_layout_dict = dict(
        height=height,
        width=width,
        margin=dict(r=0, t=0, l=0, b=0),
        mapbox_style=mapbox_style,
        mapbox_accesstoken=mapbox_accesstoken,
    )
    fig.update_layout(**update_layout_dict)
    if layout_kwargs:
        fig.update_layout(**layout_kwargs)

    update_traces_dict = dict(marker_opacity=opacity, showlegend=show_legend)
    fig.update_traces(**update_traces_dict)
    if traces_kwargs:
        fig.update_traces(**traces_kwargs)

    fig.update_coloraxes(showscale=show_legend)

    if return_plot:
        return fig
    else:
        fig.show(renderer=renderer)
        return None


def _calculate_map_centroid(regions_gdf: gpd.GeoDataFrame) -> Point:
    """
    Calculate regions centroid using Equal Area Cylindrical projection [1].

    Args:
        regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.

    Returns:
        Point: Center point in WGS84 units.

    References:
        1. https://proj.org/operations/projections/cea.html
    """
    center_point = regions_gdf.to_crs("+proj=cea").dissolve().centroid.to_crs(WGS84_CRS)[0]
    return center_point


# Inspired by:
# https://stackoverflow.com/a/65043576/7766101
def _calculate_mapbox_zoom(
    regions_gdf: gpd.GeoDataFrame,
) -> float:
    """
    Calculate approximate zoom for a plotly figure.

    Currently Plotly doesn't implement auto-fit feature for mapbox plots.

    Args:
        regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.

    Returns:
        float: zoom level for a mapbox plot.
    """
    minx, miny, maxx, maxy = regions_gdf.geometry.total_bounds
    max_bound = max(abs(maxx - minx), abs(maxy - miny)) * 111
    zoom = float(12.5 - np.log(max_bound))
    return zoom



"""GTFS2VecEmbedder tests."""

from typing import Any

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
from pytest_mock import MockerFixture
from pytorch_lightning import seed_everything
from shapely.geometry import Polygon

from srai.constants import REGIONS_INDEX
from srai.embedders import GTFS2VecEmbedder
from srai.exceptions import ModelNotFitException


@pytest.fixture  # type: ignore
def gtfs2vec_features() -> gpd.GeoDataFrame:
    """Get GTFS2Vec features GeoDataFrame."""
    features_gdf = gpd.GeoDataFrame(
        {
            "trips_at_6": [1, 0, 0],
            "trips_at_7": [1, 1, 0],
            "trips_at_8": [0, 0, 1],
            "directions_at_6": [
                {"A", "A1"},
                {"B", "B1"},
                {"C"},
            ],
        },
        geometry=gpd.points_from_xy([1, 2, 5], [1, 2, 2]),
        index=[1, 2, 3],
    )
    features_gdf.index.name = "stop_id"
    return features_gdf


@pytest.fixture  # type: ignore
def gtfs2vec_regions() -> gpd.GeoDataFrame:
    """Get GTFS2Vec regions GeoDataFrame."""
    regions_gdf = gpd.GeoDataFrame(
        {
            REGIONS_INDEX: ["ff1", "ff2", "ff3"],
        },
        geometry=[
            Polygon([(0, 0), (0, 3), (3, 3), (3, 0)]),
            Polygon([(4, 0), (4, 3), (7, 3), (7, 0)]),
            Polygon([(8, 0), (8, 3), (11, 3), (11, 0)]),
        ],
    ).set_index(REGIONS_INDEX)
    return regions_gdf


@pytest.fixture  # type: ignore
def gtfs2vec_joint() -> gpd.GeoDataFrame:
    """Get GTFS2Vec joint GeoDataFrame."""
    joint_gdf = gpd.GeoDataFrame()
    joint_gdf.index = pd.MultiIndex.from_tuples(
        [("ff1", 1), ("ff1", 2), ("ff2", 3)],
        names=[REGIONS_INDEX, "stop_id"],
    )
    return joint_gdf


@pytest.fixture  # type: ignore
def features_not_embedded() -> pd.DataFrame:
    """Get features not embedded."""
    return pd.DataFrame(
        {
            "trips_at_6": [0.5, 0.0, 0.0],
            "trips_at_7": [1.0, 0.0, 0.0],
            "trips_at_8": [0.0, 0.5, 0.0],
            "directions_at_6": [1.0, 0.25, 0.0],
            REGIONS_INDEX: ["ff1", "ff2", "ff3"],
        },
    ).set_index(REGIONS_INDEX)


@pytest.fixture  # type: ignore
def features_embedded() -> pd.DataFrame:
    """Get features embedded."""
    embeddings = np.array(
        [
            [0.642446, 0.001230, -0.038590],
            [0.960088, 0.433973, 0.401301],
            [0.224951, -0.135296, -0.157667],
            [0.782488, 0.212641, 0.177253],
        ],
        dtype=np.float32,
    )
    features = pd.DataFrame(embeddings.T)
    features.index = pd.Index(["ff1", "ff2", "ff3"], name=REGIONS_INDEX)
    features.columns = pd.RangeIndex(0, 4, 1)
    return features


@pytest.mark.parametrize(  # type: ignore
    "regions_fixture,features_fixture,joint_fixture,expectation",
    [
        (
            "gdf_unnamed_single_index",
            "gdf_features",
            "gdf_joint",
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_unnamed_single_index",
            "gdf_joint",
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_unnamed_single_index",
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_three_level_multi_index",
            pytest.raises(ValueError),
        ),
        (
            "gdf_incorrectly_named_single_index",
            "gdf_features",
            "gdf_joint",
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_incorrectly_named_single_index",
            "gdf_joint",
            pytest.raises(ValueError),
        ),
    ],
)
def test_incorrect_indexes(
    regions_fixture: str,
    features_fixture: str,
    joint_fixture: str,
    expectation: Any,
    request: Any,
) -> None:
    """Test GTFS2VecEmbedder with incorrect input dataframes."""
    regions_gdf = request.getfixturevalue(regions_fixture)
    features_gdf = request.getfixturevalue(features_fixture)
    joint_gdf = request.getfixturevalue(joint_fixture)

    with expectation:
        embedder = GTFS2VecEmbedder(skip_autoencoder=True)
        embedder.transform(regions_gdf, features_gdf, joint_gdf)

    with expectation:
        embedder = GTFS2VecEmbedder(skip_autoencoder=True)
        embedder.fit(regions_gdf, features_gdf, joint_gdf)

    with expectation:
        embedder = GTFS2VecEmbedder(skip_autoencoder=True)
        embedder.fit_transform(regions_gdf, features_gdf, joint_gdf)


def test_transform_with_unfit_model(
    gtfs2vec_regions: gpd.GeoDataFrame,
    gtfs2vec_features: gpd.GeoDataFrame,
    gtfs2vec_joint: gpd.GeoDataFrame,
) -> None:
    """Test GTFS2VecEmbedder transform with unfitted model."""
    embedder = GTFS2VecEmbedder(skip_autoencoder=False)
    with pytest.raises(ModelNotFitException):
        embedder.transform(gtfs2vec_regions, gtfs2vec_features, gtfs2vec_joint)


def test_transform_with_mismatched_features_count(
    gtfs2vec_regions: gpd.GeoDataFrame,
    gtfs2vec_features: gpd.GeoDataFrame,
    gtfs2vec_joint: gpd.GeoDataFrame,
    mocker: MockerFixture,
) -> None:
    """Test GTFS2VecEmbedder transform with mismatched features count."""
    embedder = GTFS2VecEmbedder(skip_autoencoder=False)
    mock_model = mocker.MagicMock()
    mock_model.configure_mock(n_features=42)
    embedder._model = mock_model

    with pytest.raises(ValueError):
        embedder.transform(gtfs2vec_regions, gtfs2vec_features, gtfs2vec_joint)


@pytest.mark.parametrize(  # type: ignore
    "regions_fixture,features_fixture,joint_fixture,embedding",
    [
        (
            "gtfs2vec_regions",
            "gtfs2vec_features",
            "gtfs2vec_joint",
            True,
        ),
        (
            "gtfs2vec_regions",
            "gtfs2vec_features",
            "gtfs2vec_joint",
            False,
        ),
    ],
)
def test_embedder(
    regions_fixture: str,
    features_fixture: str,
    joint_fixture: str,
    embedding: bool,
    request: Any,
) -> None:
    """Test GTFS2VecEmbedder results."""
    regions_gdf = request.getfixturevalue(regions_fixture)
    features_gdf = request.getfixturevalue(features_fixture)
    joint_gdf = request.getfixturevalue(joint_fixture)

    if embedding:
        expected_features = request.getfixturevalue("features_embedded")
    else:
        expected_features = request.getfixturevalue("features_not_embedded")

    embedder = GTFS2VecEmbedder(hidden_size=2, embedding_size=4, skip_autoencoder=not embedding)

    seed_everything(42)
    embedder.fit(regions_gdf, features_gdf, joint_gdf)
    features_embedded = embedder.transform(regions_gdf, features_gdf, joint_gdf)

    pd.testing.assert_frame_equal(features_embedded, expected_features, atol=1e-3)

    seed_everything(42)
    features_embedded = embedder.fit_transform(regions_gdf, features_gdf, joint_gdf)

    pd.testing.assert_frame_equal(features_embedded, expected_features, atol=1e-3)



"""
This module contains loaders, used to load spatial data from different sources.

We want to unify loading from different data sources into a single interface. Thanks to this, we
have a unified spatial data format, which makes it possible to feed them into any of the embedding
methods available in this library.
"""

from ._base import Loader
from .download import download_file
from .geoparquet_loader import GeoparquetLoader
from .gtfs_loader import GTFSLoader
from .osm_loaders import OSMLoader, OSMOnlineLoader, OSMPbfLoader, OSMTileLoader
from .osm_way_loader import OSMNetworkType, OSMWayLoader

__all__ = [
    "Loader",
    "GeoparquetLoader",
    "GTFSLoader",
    "OSMLoader",
    "OSMWayLoader",
    "OSMOnlineLoader",
    "OSMPbfLoader",
    "OSMTileLoader",
    "OSMNetworkType",
    "download_file",
]



"""Test case generation for Hex2VecEmbedder."""
from pathlib import Path
from typing import Optional

import geopandas as gpd
import h3
from h3ronpy.pandas.vector import cells_to_polygons
from pytorch_lightning import seed_everything

from srai.constants import REGIONS_INDEX, WGS84_CRS
from srai.embedders.hex2vec.embedder import Hex2VecEmbedder
from srai.joiners import IntersectionJoiner
from srai.loaders.osm_loaders import OSMPbfLoader
from srai.loaders.osm_loaders.filters import OsmTagsFilter
from srai.neighbourhoods import H3Neighbourhood
from srai.regionalizers import geocode_to_region_gdf
from tests.embedders.hex2vec.constants import ENCODER_SIZES, TRAINER_KWARGS


def generate_test_case(
    test_case_name: str,
    geocoding_name: str,
    root_region_index: str,
    h3_res: int,
    radius: int,
    seed: int,
    tags: Optional[OsmTagsFilter] = None,
) -> None:
    """Generate test case for Hex2VecEmbedder."""
    if tags is None:
        tags = {"leisure": "park", "amenity": "restaurant"}
    neighbourhood = H3Neighbourhood()
    regions_indexes = neighbourhood.get_neighbours_up_to_distance(root_region_index, radius)
    regions_indexes.add(root_region_index)
    regions_indexes = list(regions_indexes)  # type: ignore

    geoms = cells_to_polygons([h3.str_to_int(r) for r in regions_indexes]).values
    regions_gdf = gpd.GeoDataFrame(index=regions_indexes, geometry=geoms, crs=WGS84_CRS)
    regions_gdf.index.name = REGIONS_INDEX

    area_gdf = geocode_to_region_gdf(geocoding_name)
    loader = OSMPbfLoader()
    features_all = loader.load(area_gdf, tags=tags)

    joiner = IntersectionJoiner()
    joint_gdf = joiner.transform(regions_gdf, features_all)
    features_gdf = features_all[
        features_all.index.isin(joint_gdf.index.get_level_values("feature_id"))
    ]
    joint_gdf = joiner.transform(regions_gdf, features_gdf)

    neighbourhood = H3Neighbourhood(regions_gdf)
    seed_everything(seed)
    embedder = Hex2VecEmbedder(encoder_sizes=ENCODER_SIZES)
    results_df = embedder.fit_transform(
        regions_gdf, features_gdf, joint_gdf, neighbourhood, trainer_kwargs=TRAINER_KWARGS
    )
    results_df.columns = results_df.columns.astype(str)

    files_prefix = f"{test_case_name}"

    output_path = Path(__file__).parent / "test_files"
    regions_gdf.to_parquet(output_path / f"{files_prefix}_regions.parquet")
    features_gdf.to_parquet(output_path / f"{files_prefix}_features.parquet")
    joint_gdf.to_parquet(output_path / f"{files_prefix}_joint.parquet")
    results_df.to_parquet(output_path / f"{files_prefix}_result.parquet")



"""
Gtfs2vec embedder.

This module contains embedder from gtfs2vec paper [1].

References:
    1. https://doi.org/10.1145/3486640.3491392
"""


import json
from functools import reduce
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, Union

import geopandas as gpd
import numpy as np
import pandas as pd

from srai._optional import import_optional_dependencies
from srai.embedders import Embedder, ModelT
from srai.embedders.gtfs2vec.model import GTFS2VecModel
from srai.exceptions import ModelNotFitException
from srai.loaders.gtfs_loader import GTFS2VEC_DIRECTIONS_PREFIX, GTFS2VEC_TRIPS_PREFIX


class GTFS2VecEmbedder(Embedder):
    """GTFS2Vec Embedder."""

    def __init__(
        self, hidden_size: int = 48, embedding_size: int = 64, skip_autoencoder: bool = False
    ) -> None:
        """
        Init GTFS2VecEmbedder.

        Args:
            hidden_size (int, optional): Hidden size in encoder and decoder. Defaults to 48.
            embedding_size (int, optional): Embedding size. Defaults to 64.
            skip_autoencoder (bool, optional): Skip using autoencoder as part of embedding.
            Defaults to False.
        """
        import_optional_dependencies(
            dependency_group="torch", modules=["torch", "pytorch_lightning"]
        )
        self._model: Optional[GTFS2VecModel] = None
        self._hidden_size = hidden_size
        self._embedding_size = embedding_size
        self._skip_autoencoder = skip_autoencoder
        self._is_fitted = False

    def transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> pd.DataFrame:
        """
        Embed a given data.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.

        Returns:
            pd.DataFrame: Embedding and geometry index for each region in regions_gdf.

        Raises:
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
            ValueError: If number of features is incosistent with the model.
            ModelNotFitException: If model is not fit.
        """
        self._validate_indexes(regions_gdf, features_gdf, joint_gdf)
        features = self._prepare_features(regions_gdf, features_gdf, joint_gdf)

        if self._skip_autoencoder:
            return features
        return self._embed(features)

    def fit(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> None:
        """
        Fit model to a given data.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.

        Raises:
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
        """
        self._validate_indexes(regions_gdf, features_gdf, joint_gdf)
        features = self._prepare_features(regions_gdf, features_gdf, joint_gdf)

        if not self._skip_autoencoder:
            self._model = self._train_model_unsupervised(features)

    def fit_transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> pd.DataFrame:
        """
        Fit model and transform a given data.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.

        Returns:
            pd.DataFrame: Embedding and geometry index for each region in regions_gdf.

        Raises:
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
        """
        self._validate_indexes(regions_gdf, features_gdf, joint_gdf)
        features = self._prepare_features(regions_gdf, features_gdf, joint_gdf)

        if self._skip_autoencoder:
            return features
        else:
            self._model = self._train_model_unsupervised(features)
            return self._embed(features)

    def _maybe_get_model(self) -> GTFS2VecModel:
        """Check if model is fit and return it."""
        if self._model is None:
            raise ModelNotFitException("Model not fit! Run fit() or fit_transform() first.")
        return self._model

    def _prepare_features(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> pd.DataFrame:
        """
        Prepare features for embedding.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.
        """
        regions_gdf = self._remove_geometry_if_present(regions_gdf)
        features_gdf = self._remove_geometry_if_present(features_gdf)
        joint_gdf = self._remove_geometry_if_present(joint_gdf)

        joint_features = (
            joint_gdf.join(features_gdf, on=features_gdf.index.name)
            .groupby(regions_gdf.index.name)
            .agg(self._get_columns_aggregation(features_gdf.columns))
        )

        regions_features = (
            regions_gdf.join(joint_features, on=regions_gdf.index.name).fillna(0).astype(int)
        )
        regions_features = self._normalize_features(regions_features)
        return regions_features

    def _get_columns_aggregation(self, columns: List[str]) -> Dict[str, Any]:
        """
        Get aggregation dict for given columns.

        Args:
            columns (list): List of columns.

        Returns:
            dict: Aggregation dict.
        """
        agg_dict: Dict[str, Any] = {}

        for column in columns:
            if column.startswith(GTFS2VEC_TRIPS_PREFIX):
                agg_dict[column] = "sum"
            elif column.startswith(GTFS2VEC_DIRECTIONS_PREFIX):
                agg_dict[column] = lambda x: len(reduce(set.union, x))
        return agg_dict

    def _normalize_columns_group(self, df: pd.DataFrame, columns: List[str]) -> pd.DataFrame:
        """
        Normalize given columns in df.

        Args:
            df (pd.DataFrame): DataFrame to normalize.
            columns (list): List of columns.
        """
        df[columns] = (df[columns] - df[columns].min().min()) / (
            df[columns].max().max() - df[columns].min().min()
        )
        return df

    def _normalize_features(self, features: pd.DataFrame) -> pd.DataFrame:
        """
        Normalize features.

        Args:
            features (pd.DataFrame): Features.

        Returns:
            pd.DataFrame: Normalized features.
        """
        norm_columns = [
            [
                column
                for column in features.columns
                if column.startswith(GTFS2VEC_DIRECTIONS_PREFIX)
            ],
            [column for column in features.columns if column.startswith(GTFS2VEC_TRIPS_PREFIX)],
        ]

        for columns in norm_columns:
            features = self._normalize_columns_group(features, columns)

        return features

    def _train_model_unsupervised(self, features: pd.DataFrame) -> GTFS2VecModel:
        """
        Train model unsupervised.

        Args:
            features (pd.DataFrame): Features.
        """
        import pytorch_lightning as pl
        from torch.utils.data import DataLoader

        model = GTFS2VecModel(
            n_features=len(features.columns),
            n_hidden=self._hidden_size,
            n_embed=self._embedding_size,
        )
        X = features.to_numpy().astype(np.float32)
        x_dataloader = DataLoader(X, batch_size=24, shuffle=True, num_workers=4)
        trainer = pl.Trainer(max_epochs=10)

        trainer.fit(model, x_dataloader)

        return model

    def _embed(self, features: pd.DataFrame) -> pd.DataFrame:
        """
        Embed features.

        Args:
            features (pd.DataFrame): Features.

        Returns:
            pd.DataFrame: Embeddings.
        """
        import torch

        model = self._maybe_get_model()
        if len(features.columns) != model.n_features:
            raise ValueError(
                f"Features must have {model.n_features} columns but has {len(features.columns)}."
            )

        embeddings = model(torch.Tensor(features.to_numpy().astype(np.float32))).detach().numpy()

        return pd.DataFrame(
            data=embeddings,
            index=features.index,
        )

    def _save(self, path: Union[Path, str], embedder_config: Dict[str, Any]) -> None:
        if isinstance(path, str):
            path = Path(path)

        model = self._maybe_get_model()

        path.mkdir(parents=True, exist_ok=True)

        model.save(path / "model.pt")

        config = {
            "model_config": model.get_config(),
            "embedder_config": embedder_config,
        }
        with open(path / "config.json", "w") as f:
            json.dump(config, f, ensure_ascii=False, indent=4)

    def save(self, path: Union[Path, str]) -> None:
        """
        Save the model to a directory.

        Args:
            path (Path): Path to the directory.
        """
        embedder_config = {
            "hidden_size": self._hidden_size,
            "embedding_size": self._embedding_size,
            "skip_autoencoder": self._skip_autoencoder,
        }
        self._save(path, embedder_config)

    @classmethod
    def _load(cls, path: Union[Path, str], model_module: Type[ModelT]) -> "GTFS2VecEmbedder":
        if isinstance(path, str):
            path = Path(path)

        with (path / "config.json").open("r") as f:
            config = json.load(f)
        embedder = cls(**config["embedder_config"])
        model_path = path / "model.pt"
        model = model_module.load(model_path, **config["model_config"])
        embedder._model = model
        embedder._is_fitted = True
        return embedder

    @classmethod
    def load(cls, path: Union[Path, str]) -> "GTFS2VecEmbedder":
        """
        Load the model from a directory.

        Args:
            path (Path): Path to the directory.

        Returns:
            Hex2VecEmbedder: The loaded embedder.
        """
        return cls._load(path, GTFS2VecModel)



"""
S2 Regionalizer.

This module exposes Google's S2 Geospatial Indexing System [1] as a regionalizer.
It uses the Python API [2].

References:
    1. https://s2geometry.io/
    2. https://github.com/JoaoCarabetta/s2-py
"""

import json
from typing import Any, Dict

import geopandas as gpd
from functional import seq
from functional.pipeline import Sequence
from s2 import s2
from shapely.geometry import Polygon

from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX, WGS84_CRS
from srai.regionalizers import Regionalizer


class S2Regionalizer(Regionalizer):
    """
    S2 Regionalizer.

    S2 Regionalizer gives an opportunity to divide the given geometries into square S2 cells.
    """

    def __init__(self, resolution: int, buffer: bool = True) -> None:
        """
        Init S2 Regionalizer.

        Args:
            resolution (int): Resolution of the cells (S2 supports 0-30). See [1] for
                a full comparison.
            buffer (bool, optional): If True then fully cover geometries with S2 cells.
                Otherwise only use those cells that fully fit into them. Defaults to True.

        Raises:
            ValueError: If resolution is not between 0 and 30.

        References:
            1. https://s2geometry.io/resources/s2cell_statistics.html
        """
        if not (0 <= resolution <= 30):
            raise ValueError(f"Resolution {resolution} is not between 0 and 30.")

        self.resolution = resolution
        self.buffer = buffer

    def transform(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """
        Regionalize a given GeoDataFrame.

        Args:
            gdf (gpd.GeoDataFrame): GeoDataFrame to be regionalized.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame with regionalized geometries.
        """
        gdf_wgs84 = gdf.to_crs(crs=WGS84_CRS)

        s2_gdf = self._fill_with_s2_cells(self._explode_multipolygons(gdf_wgs84))

        # s2 library fills also holes in Polygons, so here we remove redundant cells
        res: gpd.GeoDataFrame = gpd.sjoin(
            s2_gdf,
            gdf_wgs84,
            how="inner",
            predicate="intersects" if self.buffer else "within",
        ).drop(columns=["index_right"])

        res = res[~res.index.duplicated(keep="first")]

        res.index.name = REGIONS_INDEX

        return res

    def _fill_with_s2_cells(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        geo_json = json.loads(gdf.to_json())
        cells = (
            seq(geo_json["features"])
            .flat_map(lambda f: self._geojson_to_cells(f[GEOMETRY_COLUMN], self.resolution))
            .to_dict()
        )
        cells_gdf = gpd.GeoDataFrame(
            None,
            index=cells.keys(),
            geometry=list(cells.values()),
            crs=WGS84_CRS,
        )

        return cells_gdf

    def _geojson_to_cells(self, geo_json: Dict[str, Any], res: int) -> Sequence:
        raw_cells = s2.polyfill(geo_json, res, with_id=True, geo_json_conformant=True)
        cells: Sequence = seq(raw_cells).map(lambda c: (c["id"], Polygon(c[GEOMETRY_COLUMN])))

        return cells



"""
This module contains regionalizers, which are used to divide space before analysis.

Embedding methods available in `srai` operate on a regions, which can be defined in many ways. In
this module, we aggregate different regionalization methods under a common `Regionalizer` interface.
We include both pre-defined spatial indexes (e.g. H3 or S2), data-based ones (e.g. Voronoi) and OSM-
based ones (e.g. administrative boundaries).
"""

from ._base import Regionalizer
from .administrative_boundary_regionalizer import AdministrativeBoundaryRegionalizer
from .geocode import geocode_to_region_gdf
from .h3_regionalizer import H3Regionalizer
from .s2_regionalizer import S2Regionalizer
from .slippy_map_regionalizer import SlippyMapRegionalizer
from .voronoi_regionalizer import VoronoiRegionalizer

__all__ = [
    "Regionalizer",
    "AdministrativeBoundaryRegionalizer",
    "H3Regionalizer",
    "S2Regionalizer",
    "VoronoiRegionalizer",
    "SlippyMapRegionalizer",
    "geocode_to_region_gdf",
]



"""GTFS Loader tests."""
from pathlib import Path
from typing import Any
from unittest import TestCase

import pandas as pd
import pytest
from pytest_mock import MockerFixture

from srai.constants import GEOMETRY_COLUMN
from srai.loaders import GTFSLoader
from srai.loaders.gtfs_loader import GTFS2VEC_DIRECTIONS_PREFIX, GTFS2VEC_TRIPS_PREFIX

ut = TestCase()


def test_validation_ok(mocker: MockerFixture, gtfs_validation_ok: pd.DataFrame) -> None:
    """Test checks if GTFSLoader returns no errors."""
    feed_mock = mocker.MagicMock()
    feed_mock.configure_mock(**{"validate.return_value": gtfs_validation_ok})

    loader = GTFSLoader()
    loader._validate_feed(feed_mock)


def test_validation_error(mocker: MockerFixture, gtfs_validation_error: pd.DataFrame) -> None:
    """Test checks if GTFSLoader raises ValueError on validation error."""
    feed_mock = mocker.MagicMock()
    feed_mock.configure_mock(**{"validate.return_value": gtfs_validation_error})

    warning_mock = mocker.patch("warnings.warn")

    loader = GTFSLoader()
    with pytest.raises(ValueError):
        loader._validate_feed(feed_mock)
        warning_mock.assert_called_once()


def test_validation_warning(mocker: MockerFixture, gtfs_validation_error: pd.DataFrame) -> None:
    """Test checks if GTFSLoader raises ValueError on validation error."""
    feed_mock = mocker.MagicMock()
    feed_mock.configure_mock(**{"validate.return_value": gtfs_validation_error})

    warning_mock = mocker.patch("warnings.warn")

    loader = GTFSLoader()
    loader._validate_feed(feed_mock, fail=False)
    warning_mock.assert_called_once()


def test_gtfs_loader(feed: Any, mocker: MockerFixture, gtfs_validation_ok: pd.DataFrame) -> None:
    """Test GTFSLoader."""
    feed.validate.return_value = gtfs_validation_ok
    mocker.patch("gtfs_kit.read_feed", return_value=feed)

    loader = GTFSLoader()
    features = loader.load(Path("feed.zip").resolve())

    ut.assertCountEqual(features.index, ["42", "76"])
    ut.assertCountEqual(
        features.columns,
        [
            f"{GTFS2VEC_TRIPS_PREFIX}12",
            f"{GTFS2VEC_TRIPS_PREFIX}13",
            f"{GTFS2VEC_DIRECTIONS_PREFIX}12",
            f"{GTFS2VEC_DIRECTIONS_PREFIX}13",
            GEOMETRY_COLUMN,
        ],
    )


def test_gtfs_loader_with_invalid_feed(
    feed: Any, mocker: MockerFixture, gtfs_validation_error: pd.DataFrame
) -> None:
    """Test GTFSLoader with invalid feed."""
    feed.validate.return_value = gtfs_validation_error
    mocker.patch("gtfs_kit.read_feed", return_value=feed)
    warning_mock = mocker.patch("warnings.warn")

    loader = GTFSLoader()
    with pytest.raises(ValueError):
        loader.load(Path("feed.zip").resolve())
        warning_mock.assert_called_once()


def test_gtfs_loader_skip_validation(
    feed: Any, mocker: MockerFixture, gtfs_validation_ok: pd.DataFrame
) -> None:
    """Test GTFSLoader with invalid feed."""
    feed.validate.return_value = gtfs_validation_ok
    mocker.patch("gtfs_kit.read_feed", return_value=feed)

    loader = GTFSLoader()
    loader.load(Path("feed.zip").resolve(), skip_validation=True)

    feed.validate.assert_not_called()



"""
This module contains plotting methods.

We provide some high-level plotting methods which work on the outputs of different `srai`
components. By default, `folium` based functions are exposed within `plotting` module. Additional
functions can be found in `srai.plotting.plotly_wrapper` module.
"""
from .folium_wrapper import plot_all_neighbourhood, plot_neighbours, plot_numeric_data, plot_regions

__all__ = ["plot_regions", "plot_numeric_data", "plot_neighbours", "plot_all_neighbourhood"]



"""H3 IJ coordinates tests."""

from typing import List

import h3
import numpy as np
import pytest

from srai.h3 import get_local_ij_index


@pytest.mark.parametrize(
    "h3_origin",
    [
        "891e2040d4bffff",
        "871e20400ffffff",
        "821f77fffffffff",
        "81743ffffffffff",
    ],
)  # type: ignore
def test_self_ok(h3_origin: str) -> None:
    """Test checks if self coordinates are in origin."""
    coordinates = get_local_ij_index(origin_index=h3_origin, h3_index=h3_origin)
    assert coordinates == (0, 0)


@pytest.mark.parametrize(
    "h3_origin, h3_cell",
    [
        ("871f53c93ffffff", "871f53c91ffffff"),
        ("861fae207ffffff", "861fae22fffffff"),
        ("81597ffffffffff", "813fbffffffffff"),
        ("84be185ffffffff", "84be181ffffffff"),
    ],
)  # type: ignore
def test_string_ok(h3_origin: str, h3_cell: str) -> None:
    """Test checks if pairs are in right orientation."""
    coordinates = get_local_ij_index(origin_index=h3_origin, h3_index=h3_cell)
    assert coordinates == (0, 1)


@pytest.mark.parametrize(
    "h3_origin, h3_cells",
    [
        (
            "892a100d6d3ffff",
            [
                "892a100896fffff",
                "892a100d6d7ffff",
                "892a100d6c3ffff",
                "892a100d6dbffff",
                "892a1008ba7ffff",
                "892a100896bffff",
            ],
        ),
        (
            "86195da4fffffff",
            [
                "86194ad37ffffff",
                "86194ad17ffffff",
                "86194ada7ffffff",
                "86195da5fffffff",
                "86195da47ffffff",
                "86195da6fffffff",
            ],
        ),
        (
            "8a1e24aa5637fff",
            [
                "8a1e24aa5627fff",
                "8a1e24aa5607fff",
                "8a1e24aa5617fff",
                "8a1e24aa578ffff",
                "8a1e24aa57affff",
                "8a1e24aa571ffff",
            ],
        ),
    ],
)  # type: ignore
@pytest.mark.parametrize("return_as_numpy", [True, False])  # type: ignore
def test_list_ok(h3_origin: str, h3_cells: List[str], return_as_numpy: bool) -> None:
    """Test checks if lists are parsed correctly."""
    coordinates = get_local_ij_index(
        origin_index=h3_origin, h3_index=h3_cells, return_as_numpy=return_as_numpy
    )

    expected_coordinates = [(0, 1), (1, 1), (1, 0), (0, -1), (-1, -1), (-1, 0)]

    if return_as_numpy:
        np.testing.assert_array_equal(coordinates, expected_coordinates)
    else:
        assert coordinates == expected_coordinates


@pytest.mark.parametrize(
    "h3_origin, h3_cell",
    [
        ("83a75dfffffffff", "83a791fffffffff"),
        ("84a605bffffffff", "84a6021ffffffff"),
        ("836200fffffffff", "837400fffffffff"),
    ],
)  # type: ignore
def test_pentagon_error(h3_origin: str, h3_cell: str) -> None:
    """Test checks if method fails over pentagon pairs."""
    with pytest.raises(h3._cy.error_system.H3FailedError):
        get_local_ij_index(origin_index=h3_origin, h3_index=h3_cell)



"""
This module contains embedders, used to convert spatial data to their vector representations.

Embedders are designed to unify different types of spatial data embedding methods, such as hex2vec
or gtfs2vec into a single interface. This allows to easily switch between different embedding
methods without changing the rest of the code.
"""


from ._base import Embedder, Model, ModelT
from .contextual_count_embedder import ContextualCountEmbedder
from .count_embedder import CountEmbedder
from .gtfs2vec import GTFS2VecEmbedder
from .hex2vec import Hex2VecEmbedder
from .highway2vec import Highway2VecEmbedder

__all__ = [
    "Embedder",
    "CountEmbedder",
    "ContextualCountEmbedder",
    "GTFS2VecEmbedder",
    "Hex2VecEmbedder",
    "Highway2VecEmbedder",
    "Model",
    "ModelT",
]



"""
Slippy map regionalizer.

This module implements Slippy map tilenames [1] as a regionalizer.

References:
    1. https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
"""
from itertools import product
from typing import Any, Dict, List, Tuple

import geopandas as gpd
import numpy as np
import shapely.geometry as shpg
from functional import seq

from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX, WGS84_CRS
from srai.regionalizers import Regionalizer


class SlippyMapRegionalizer(Regionalizer):
    """SlippyMapRegionalizer class."""

    def __init__(self, zoom: int) -> None:
        """
        Initialize SlippyMapRegionalizer.

        Args:
            zoom (int): zoom level

        Raises:
            ValueError: if zoom is not in [0, 19]
        """
        if not 0 <= zoom <= 19:
            raise ValueError
        self.zoom = zoom
        super().__init__()

    def transform(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """
        Regionalize a given GeoDataFrame.

        Args:
            gdf (gpd.GeoDataFrame): GeoDataFrame to be regionalized.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame with regionalized geometries.

        Raises:
            ValueError: If provided GeoDataFrame has no crs defined.
        """
        gdf_wgs84 = gdf.to_crs(crs=WGS84_CRS)
        gdf_exploded = self._explode_multipolygons(gdf_wgs84)

        values = (
            seq(gdf_exploded[GEOMETRY_COLUMN])
            .map(self._to_cells)
            .flat_map(lambda x: x)
            .map(
                lambda item: {
                    **item,
                    REGIONS_INDEX: f"{item['x']}_{item['y']}_{self.zoom}",
                    "z": self.zoom,
                }
            )
            .to_list()
        )

        gdf = gpd.GeoDataFrame(values, geometry=GEOMETRY_COLUMN, crs=WGS84_CRS).set_index(
            REGIONS_INDEX
        )
        return gdf.drop_duplicates()

    def _to_cells(self, polygon: shpg.Polygon) -> List[Dict[str, Any]]:
        gdf_bounds = polygon.bounds
        x_start, y_start = self._coordinates_to_x_y(gdf_bounds[1], gdf_bounds[0])
        x_end, y_end = self._coordinates_to_x_y(gdf_bounds[3], gdf_bounds[2])
        tiles = []
        for x, y in product(range(x_start, x_end + 1), range(y_end, y_start + 1)):
            tile_polygon = self._polygon_from_x_y(x, y)
            if not self._should_skip_tile(polygon, tile_polygon):
                tiles.append(dict(x=x, y=y, geometry=tile_polygon))
        return tiles

    def _polygon_from_x_y(self, x: int, y: int) -> shpg.Polygon:
        latitude_start, longitude_start = self._x_y_to_coordinates(x, y)
        latitude_end, longitude_end = self._x_y_to_coordinates(x + 1, y + 1)
        tile_polygon = shpg.box(
            minx=longitude_start, miny=latitude_start, maxx=longitude_end, maxy=latitude_end
        )
        return tile_polygon

    def _should_skip_tile(self, area: shpg.Polygon, tile: shpg.Polygon) -> bool:
        """
        Check if tile is outside area boundaries.

        If so, skip it.
        """
        intersects: bool = tile.intersects(area)
        return not intersects

    def _coordinates_to_x_y(self, latitude: float, longitude: float) -> Tuple[int, int]:
        """
        Convert latitude and longitude into x and y values using `self.zoom`.

        Based on https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Implementations.
        """
        n_rows = 2**self.zoom
        x_tile = int(n_rows * ((longitude + 180) / 360))
        lat_radian = np.radians(latitude)
        y_tile = int((1 - np.arcsinh(np.tan(lat_radian)) / np.pi) / 2 * n_rows)
        return x_tile, y_tile

    def _x_y_to_coordinates(self, x: int, y: int) -> Tuple[float, float]:
        """
        Convert x and y values into latitude and longitude using `self.zoom`.

        Based on https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Implementations.
        """
        n = 2.0**self.zoom
        lon_deg = x / n * 360.0 - 180.0
        lat_rad = np.arctan(np.sinh(np.pi * (1 - 2 * y / n)))
        lat_deg = np.degrees(lat_rad)
        return (lat_deg, lon_deg)



"""Tests for merging OSM Loaders filters."""
from contextlib import nullcontext as does_not_raise
from typing import Any
from unittest import TestCase

import pytest

from srai.loaders.osm_loaders.filters import OsmTagsFilter, merge_grouped_osm_tags_filter

ut = TestCase()


@pytest.mark.parametrize(  # type: ignore
    "grouped_filter,expected_result_filter,expectation",
    [
        ({"tag_a": True}, {"tag_a": True}, pytest.raises(ValueError)),
        ({"tag_a": "A"}, {"tag_a": "A"}, pytest.raises(ValueError)),
        ({"tag_a": ["A"]}, {"tag_a": ["A"]}, pytest.raises(ValueError)),
        ({}, {}, does_not_raise()),
        ({"group_a": {}}, {}, does_not_raise()),
        ({"group_a": {"tag_a": True}}, {"tag_a": True}, does_not_raise()),
        ({"group_a": {"tag_a": "A"}}, {"tag_a": ["A"]}, does_not_raise()),
        ({"group_a": {"tag_a": ["A"]}}, {"tag_a": ["A"]}, does_not_raise()),
        (
            {"group_a": {"tag_a": "A", "tag_b": "B"}},
            {"tag_a": ["A"], "tag_b": ["B"]},
            does_not_raise(),
        ),
        (
            {"group_a": {"tag_a": ["A"], "tag_b": ["B"]}},
            {"tag_a": ["A"], "tag_b": ["B"]},
            does_not_raise(),
        ),
        (
            {
                "group_a": {"tag_a": "A", "tag_b": "B"},
                "group_b": {"tag_a": "A", "tag_b": "B"},
            },
            {"tag_a": ["A"], "tag_b": ["B"]},
            does_not_raise(),
        ),
        (
            {
                "group_a": {"tag_a": "A", "tag_b": "B"},
                "group_b": {"tag_c": "C", "tag_d": "D"},
            },
            {"tag_a": ["A"], "tag_b": ["B"], "tag_c": ["C"], "tag_d": ["D"]},
            does_not_raise(),
        ),
        (
            {
                "group_a": {"tag_a": "A", "tag_b": "B"},
                "group_b": {"tag_a": "C", "tag_b": "D"},
            },
            {"tag_a": ["A", "C"], "tag_b": ["B", "D"]},
            does_not_raise(),
        ),
        (
            {
                "group_a": {"tag_a": "A", "tag_b": "B"},
                "group_b": {"tag_a": ["C", "D"], "tag_b": "E"},
            },
            {"tag_a": ["A", "C", "D"], "tag_b": ["B", "E"]},
            does_not_raise(),
        ),
        (
            {
                "group_a": {"tag_a": "A", "tag_b": "B"},
                "group_b": {"tag_a": ["C", "D"], "tag_b": True},
            },
            {"tag_a": ["A", "C", "D"], "tag_b": True},
            does_not_raise(),
        ),
        (
            {
                "group_a": {"tag_a": ["A"], "tag_b": ["B"]},
                "group_b": {"tag_a": ["C", "D"], "tag_b": False},
            },
            {"tag_a": ["A", "C", "D"], "tag_b": ["B"]},
            does_not_raise(),
        ),
        (
            {
                "group_a": {"tag_a": ["A", "C"], "tag_b": ["B", "E"]},
                "group_b": {"tag_a": ["C", "D"], "tag_b": ["B"]},
            },
            {"tag_a": ["A", "C", "D"], "tag_b": ["B", "E"]},
            does_not_raise(),
        ),
    ],
)
def test_merge_grouped_filters(
    grouped_filter: Any, expected_result_filter: OsmTagsFilter, expectation: Any
) -> None:
    """Test merging grouped tags filter into a base osm filter."""
    with expectation:
        merged_filters = merge_grouped_osm_tags_filter(grouped_filter)
        ut.assertDictEqual(expected_result_filter, merged_filters)



"""Tests for H3Regionalizer."""
from contextlib import nullcontext as does_not_raise
from typing import TYPE_CHECKING, Any, List
from unittest import TestCase

import pytest

from srai.constants import GEOMETRY_COLUMN
from srai.regionalizers import H3Regionalizer
from srai.regionalizers.geocode import geocode_to_region_gdf

if TYPE_CHECKING:
    import geopandas as gpd


ut = TestCase()
H3_RESOLUTION = 3


@pytest.fixture  # type: ignore
def expected_h3_indexes() -> List[str]:
    """Get expected h3 indexes."""
    return [
        "837559fffffffff",
        "83754efffffffff",
        "83754cfffffffff",
        "837541fffffffff",
        "83755dfffffffff",
        "837543fffffffff",
        "83754afffffffff",
    ]


@pytest.fixture  # type: ignore
def expected_unbuffered_h3_indexes() -> List[str]:
    """Get expected h3 index for the unbuffered case."""
    return [
        "83754efffffffff",
    ]


@pytest.mark.parametrize(  # type: ignore
    "gdf_fixture,expected_h3_indexes_fixture,resolution,buffer,expectation",
    [
        ("gdf_polygons", "expected_h3_indexes", H3_RESOLUTION, True, does_not_raise()),
        ("gdf_polygons", "expected_unbuffered_h3_indexes", H3_RESOLUTION, False, does_not_raise()),
        ("gdf_multipolygon", "expected_h3_indexes", H3_RESOLUTION, True, does_not_raise()),
        ("gdf_empty", "expected_h3_indexes", H3_RESOLUTION, True, pytest.raises(AttributeError)),
        ("gdf_polygons", "expected_h3_indexes", -1, True, pytest.raises(ValueError)),
        ("gdf_polygons", "expected_h3_indexes", 16, True, pytest.raises(ValueError)),
        ("gdf_no_crs", "expected_h3_indexes", H3_RESOLUTION, True, pytest.raises(ValueError)),
    ],
)
def test_transform(
    gdf_fixture: str,
    expected_h3_indexes_fixture: str,
    resolution: int,
    buffer: bool,
    expectation: Any,
    request: Any,
) -> None:
    """Test transform of H3Regionalizer."""
    gdf: gpd.GeoDataFrame = request.getfixturevalue(gdf_fixture)
    h3_indexes: List[str] = request.getfixturevalue(expected_h3_indexes_fixture)
    with expectation:
        gdf_h3 = H3Regionalizer(resolution, buffer=buffer).transform(gdf)

        ut.assertCountEqual(first=gdf_h3.index.to_list(), second=h3_indexes)
        assert GEOMETRY_COLUMN in gdf_h3


def test_wroclaw_edge_case() -> None:
    """Test edge case from H3Neighbourhood example error."""
    gdf_wro = geocode_to_region_gdf("Wrocław, PL")
    regions_gdf = H3Regionalizer(8).transform(gdf_wro)

    edge_region_id = "881e2050bdfffff"

    assert edge_region_id in regions_gdf.index, "Edge cell is not in the regions."



"""Tests for TileLoader class."""
from io import BytesIO
from typing import List
from urllib.parse import urljoin

import geopandas as gpd
import pytest
import requests_mock
from functional import seq
from numpy.random import default_rng
from PIL import Image
from shapely.geometry import Polygon

from srai.constants import WGS84_CRS
from srai.loaders.osm_loaders import OSMTileLoader

TEST_DOMAIN = "http://mock_server"
RESOURCE_TYPE = "png"
LOCATION = "Wroclaw, Poland"
SAVE_TO_DISK_DIR = "123"
ZOOM = 10

rng = default_rng()


def to_bytes(img: Image.Image) -> bytes:
    """Convert image into bytes."""
    img_bytes = BytesIO()
    img.save(img_bytes, RESOURCE_TYPE)
    return img_bytes.getvalue()


@pytest.fixture  # type: ignore
def loader() -> OSMTileLoader:
    """Create default TileLoader object."""
    return OSMTileLoader(TEST_DOMAIN, zoom=ZOOM, verbose=False)


@pytest.fixture  # type: ignore
def images() -> List[bytes]:
    """Create list of images as bytes."""
    return [
        to_bytes(Image.fromarray(rng.integers(low=0, high=256, size=(4, 4, 3), dtype="uint8")))
        for _ in range(3)
    ]


@pytest.fixture  # type: ignore
def gdf() -> gpd.GeoDataFrame:
    """GeoDataFrame approximating Wroclaw bounds."""
    polygon = Polygon(
        [
            (16.8073393, 51.1389477),
            (17.0278673, 51.0426754),
            (17.1762192, 51.1063195),
            (16.9580276, 51.2093551),
            (16.8073393, 51.1389477),
        ]
    )
    gdf = gpd.GeoDataFrame({"geometry": [polygon]}, crs=WGS84_CRS)
    return gdf


def mock_requests(images: List[bytes], m: requests_mock.Mocker) -> None:
    """Make mocks for requests."""
    m.get(urljoin(TEST_DOMAIN, f"10/560/341.{RESOURCE_TYPE}"), content=images[0])
    m.get(urljoin(TEST_DOMAIN, f"10/559/342.{RESOURCE_TYPE}"), content=images[1])
    m.get(urljoin(TEST_DOMAIN, f"10/560/342.{RESOURCE_TYPE}"), content=images[2])


def test_load_images_properly(
    images: List[bytes],
    gdf: gpd.GeoDataFrame,
    loader: OSMTileLoader,
) -> None:
    """Test if load returns proper images list according to location."""
    with requests_mock.Mocker() as m:
        mock_requests(images, m)
        tiles = loader.load(gdf)

        assert len(m.request_history) == 3, f"Got {len(m.request_history)} requests."

    assert to_bytes(tiles.loc[f"560_341_{ZOOM}", "tile"]) == images[0]
    assert to_bytes(tiles.loc[f"559_342_{ZOOM}", "tile"]) == images[1]
    assert to_bytes(tiles.loc[f"560_342_{ZOOM}", "tile"]) == images[2]

    xs_by_index = seq(tiles.index).map(lambda x: int(x[:3])).to_list()
    ys_by_index = seq(tiles.index).map(lambda x: int(x[4:7])).to_list()
    zs_by_index = seq(tiles.index).map(lambda x: int(x[8:])).to_list()

    assert xs_by_index == tiles["x"].tolist()
    assert ys_by_index == tiles["y"].tolist()
    assert zs_by_index == tiles["z"].tolist()


def test_should_throw_with_save_and_not_path() -> None:
    """Test checking if throws on none path with save strategy."""
    with pytest.raises(ValueError):
        _ = OSMTileLoader(
            tile_server_url=TEST_DOMAIN, zoom=ZOOM, data_collector="save", storage_path=None
        )



"""Tests for OSMPbfLoader."""
from pathlib import Path
from typing import List, Union
from unittest import TestCase

import geopandas as gpd
import pytest
from shapely.geometry import LineString, MultiPolygon, Point, Polygon
from shapely.geometry.base import BaseGeometry

from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX, WGS84_CRS
from srai.loaders.osm_loaders import OSMPbfLoader
from srai.loaders.osm_loaders.filters import (
    BASE_OSM_GROUPS_FILTER,
    HEX2VEC_FILTER,
    GroupedOsmTagsFilter,
    OsmTagsFilter,
)
from srai.loaders.osm_loaders.pbf_file_downloader import PbfFileDownloader, PbfSourceLiteral
from srai.loaders.osm_loaders.pbf_file_handler import PbfFileHandler

ut = TestCase()


@pytest.mark.parametrize(  # type: ignore
    "test_polygon",
    [
        (Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])),
        (Polygon([(3, 5), (3, 10), (7, 10), (7, 5)])),
        (Polygon([(-30, -30), (-30, 30), (30, 30), (30, -30)])),
        (
            Polygon(
                shell=[
                    (-1, 0),
                    (0, 0.5),
                    (1, 0),
                    (1, 1),
                    (0, 1),
                ],
                holes=[
                    [
                        (0.8, 0.9),
                        (0.9, 0.55),
                        (0.8, 0.3),
                        (0.5, 0.4),
                    ]
                ],
            )
        ),
        (
            gpd.read_file(
                filename=Path(__file__).parent / "test_files" / "poland.geojson"
            ).geometry[0]
        ),
        (
            gpd.read_file(
                filename=Path(__file__).parent
                / "test_files"
                / "south_africa_without_islands.geojson"
            ).geometry[0]
        ),
    ],
)
def test_geometry_preparing(test_polygon: BaseGeometry):
    """Test proper geometry preparing in `PbfFileDownloader`."""
    downloader = PbfFileDownloader(
        download_source="protomaps",
        switch_to_geofabrik_on_error=False,
    )
    prepared_polygon = downloader._prepare_polygon_for_download(test_polygon)

    assert len(prepared_polygon.exterior.coords) <= 1000
    assert len(prepared_polygon.interiors) == 0
    assert test_polygon.covered_by(prepared_polygon)


@pytest.mark.parametrize(
    "test_geometry", [Point(1, 1), LineString([(1, 1), (0, 1)])]
)  # type: ignore
@pytest.mark.parametrize(
    "pbf_source", ["geofabrik", "openstreetmap_fr", "protomaps"]
)  # type: ignore
def test_disallow_non_polygons(test_geometry: BaseGeometry, pbf_source: PbfSourceLiteral) -> None:
    """Test if `PbfFileDownloader` disallows non polygon geometries."""
    with pytest.raises(ValueError):
        regions_gdf = gpd.GeoDataFrame(
            geometry=[test_geometry],
            index=gpd.pd.Index(name=REGIONS_INDEX, data=[1]),
            crs=WGS84_CRS,
        )
        downloader = PbfFileDownloader(
            download_source=pbf_source,
            switch_to_geofabrik_on_error=False,
        )
        downloader.download_pbf_files_for_regions_gdf(regions_gdf)


@pytest.mark.parametrize(  # type: ignore
    "test_polygon,test_file_names",
    [
        (
            Polygon([(0, 0), (0, 1), (1, 1), (1, 0)]),
            ["eb2848d259345ce7dfe8af34fd1ab24503bb0b952e04e872c87c55550fa50fbf.osm.pbf"],
        ),
        (
            Polygon(
                shell=[
                    (-1, 0),
                    (0, 0.5),
                    (1, 0),
                    (1, 1),
                    (0, 1),
                ],
                holes=[
                    [
                        (0.8, 0.9),
                        (0.9, 0.55),
                        (0.8, 0.3),
                        (0.5, 0.4),
                    ]
                ],
            ),
            ["529cdcbb7a3cc103658ef31b39bed24984e421127d319c867edf2f86ff3bb098.osm.pbf"],
        ),
        (
            MultiPolygon(
                [
                    (
                        [
                            (-13, -8),
                            (-13, -9),
                            (-12, -9),
                            (-12, -8),
                        ],
                        (),
                    ),
                    (
                        [(-0.25, 0), (0.25, 0), (0, 0.2)],
                        (),
                    ),
                ]
            ),
            [
                "7a0163cb721992d6219d486b3d29517d06aa0db19dd7be049f4f1fabf6146073.osm.pbf",
                "aa756ad3a961ba6d9da46c712b0d979d0c7d4768641ceea7409b287e2d18a48f.osm.pbf",
            ],
        ),
    ],
)
def test_pbf_downloading(test_polygon: BaseGeometry, test_file_names: List[str]):
    """Test proper files downloading in `PbfFileDownloader`."""
    regions_gdf = gpd.GeoDataFrame(
        geometry=[test_polygon],
        index=gpd.pd.Index(name=REGIONS_INDEX, data=[1]),
        crs=WGS84_CRS,
    )
    downloader = PbfFileDownloader(
        download_source="protomaps",
        download_directory=Path(__file__).parent / "test_files",
        switch_to_geofabrik_on_error=False,
    )
    files = downloader.download_pbf_files_for_regions_gdf(regions_gdf)
    file_names = [path.name for path in files[1]]
    assert set(file_names) == set(test_file_names)


@pytest.mark.parametrize(  # type: ignore
    "test_file_name,query,expected_result_length,expected_features_columns_length",
    [
        (
            "d17f922ed15e9609013a6b895e1e7af2d49158f03586f2c675d17b760af3452e.osm.pbf",
            None,
            678,
            274,
        ),
        (
            "eb2848d259345ce7dfe8af34fd1ab24503bb0b952e04e872c87c55550fa50fbf.osm.pbf",
            None,
            1,
            23,
        ),
        ("529cdcbb7a3cc103658ef31b39bed24984e421127d319c867edf2f86ff3bb098.osm.pbf", None, 0, 0),
        (
            "d17f922ed15e9609013a6b895e1e7af2d49158f03586f2c675d17b760af3452e.osm.pbf",
            HEX2VEC_FILTER,
            97,
            10,
        ),
        (
            "eb2848d259345ce7dfe8af34fd1ab24503bb0b952e04e872c87c55550fa50fbf.osm.pbf",
            HEX2VEC_FILTER,
            0,
            0,
        ),
    ],
)
def test_pbf_handler(
    test_file_name: str,
    query: OsmTagsFilter,
    expected_result_length: int,
    expected_features_columns_length: int,
):
    """Test proper files loading in `PbfFileHandler`."""
    handler = PbfFileHandler(tags=query)
    features_gdf = handler.get_features_gdf(
        file_paths=[Path(__file__).parent / "test_files" / test_file_name]
    )
    assert len(features_gdf) == expected_result_length
    assert len(features_gdf.columns) == expected_features_columns_length + 1


def test_pbf_handler_geometry_filtering():  # type: ignore
    """Test proper spatial data filtering in `PbfFileHandler`."""
    file_name = "d17f922ed15e9609013a6b895e1e7af2d49158f03586f2c675d17b760af3452e.osm.pbf"
    handler = PbfFileHandler(
        tags=HEX2VEC_FILTER, region_geometry=Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])
    )
    features_gdf = handler.get_features_gdf(
        file_paths=[Path(__file__).parent / "test_files" / file_name]
    )
    assert len(features_gdf) == 0


@pytest.mark.parametrize(  # type: ignore
    "test_geometries,pbf_file,query,pbf_source,expected_result_length,expected_features_columns_length,expected_features_columns_names",
    [
        (
            [
                Polygon(
                    [
                        (7.416769421059001, 43.7346112362936),
                        (7.416769421059001, 43.730681304758946),
                        (7.4218262821731, 43.730681304758946),
                        (7.4218262821731, 43.7346112362936),
                    ]
                )
            ],
            None,
            HEX2VEC_FILTER,
            "geofabrik",
            397,
            12,
            [
                "amenity",
                "building",
                "healthcare",
                "historic",
                "landuse",
                "leisure",
                "natural",
                "office",
                "shop",
                "sport",
                "tourism",
                "water",
            ],
        ),
        (
            [Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])],
            Path(__file__).parent
            / "test_files"
            / "d17f922ed15e9609013a6b895e1e7af2d49158f03586f2c675d17b760af3452e.osm.pbf",
            HEX2VEC_FILTER,
            "protomaps",
            0,
            0,
            [],
        ),
        (
            [Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])],
            Path(__file__).parent
            / "test_files"
            / "d17f922ed15e9609013a6b895e1e7af2d49158f03586f2c675d17b760af3452e.osm.pbf",
            BASE_OSM_GROUPS_FILTER,
            "protomaps",
            0,
            0,
            [],
        ),
    ],
)
def test_osm_pbf_loader(
    test_geometries: List[BaseGeometry],
    pbf_file: Path,
    query: Union[OsmTagsFilter, GroupedOsmTagsFilter],
    pbf_source: PbfSourceLiteral,
    expected_result_length: int,
    expected_features_columns_length: int,
    expected_features_columns_names: List[str],
):
    """Test `OSMPbfLoader.load()`."""
    download_directory = Path(__file__).parent / "test_files"
    area = gpd.GeoDataFrame(
        geometry=test_geometries,
        index=gpd.pd.Index(name=REGIONS_INDEX, data=list(range(len(test_geometries)))),
        crs=WGS84_CRS,
    )

    loader = OSMPbfLoader(
        pbf_file=pbf_file,
        download_directory=download_directory,
        download_source=pbf_source,
        switch_to_geofabrik_on_error=False,
    )
    result = loader.load(area, tags=query)

    assert (
        len(result) == expected_result_length
    ), f"Mismatched result length ({len(result)}, {expected_result_length})"
    assert (
        len(result.columns) == expected_features_columns_length + 1
    ), f"Mismatched columns length ({len(result.columns)}, {expected_features_columns_length + 1})"
    ut.assertCountEqual(result.columns, expected_features_columns_names + [GEOMETRY_COLUMN])



"""Tests for OSMOnlineLoader."""
from typing import TYPE_CHECKING, Any

import geopandas as gpd
import osmnx as ox
import pytest
from packaging import version
from pandas.testing import assert_frame_equal

from srai.constants import WGS84_CRS
from srai.loaders.osm_loaders import OSMOnlineLoader
from srai.loaders.osm_loaders.filters import OsmTagsFilter

if TYPE_CHECKING:
    from shapely.geometry import Polygon


@pytest.fixture  # type: ignore
def mock_osmnx(
    mocker, two_polygons_area_gdf, area_with_no_objects_gdf, amenities_gdf, building_gdf
):
    """Patch `osmnx` functions to return data from predefined gdfs."""
    gdfs = {"amenity": amenities_gdf, "building": building_gdf}
    polygon_1, polygon_2 = two_polygons_area_gdf["geometry"]
    empty_polygon = area_with_no_objects_gdf["geometry"][0]

    def mock_geometries_from_polygon(polygon: "Polygon", tags: OsmTagsFilter) -> gpd.GeoDataFrame:
        tag_key, tag_value = list(tags.items())[0]
        gdf = gdfs[tag_key]
        if tag_value is True:
            tag_res = gdf
        else:
            tag_res = gdf.loc[gdf[tag_key] == tag_value]
        if tag_res.empty:
            return tag_res
        if polygon == polygon_1:
            return tag_res.iloc[:1]
        elif polygon == polygon_2:
            return tag_res.iloc[1:]
        elif polygon == empty_polygon:
            return gpd.GeoDataFrame(crs=WGS84_CRS, geometry=[])
        return None

    if version.parse(ox.__version__) >= version.parse("1.5.0"):
        mocker.patch("osmnx.features_from_polygon", new=mock_geometries_from_polygon)
    else:
        mocker.patch("osmnx.geometries_from_polygon", new=mock_geometries_from_polygon)


@pytest.mark.parametrize(  # type: ignore
    "area_gdf_fixture,query,expected_result_gdf_fixture",
    [
        ("single_polygon_area_gdf", {"amenity": "restaurant"}, "expected_result_single_polygon"),
        ("two_polygons_area_gdf", {"amenity": "restaurant"}, "expected_result_gdf_simple"),
        (
            "two_polygons_area_gdf",
            {"amenity": ["restaurant", "bar"], "building": True},
            "expected_result_gdf_complex",
        ),
        (
            "empty_area_gdf",
            {"amenity": ["restaurant", "bar"], "building": True},
            "empty_result_gdf",
        ),
        (
            "area_with_no_objects_gdf",
            {"amenity": ["restaurant", "bar"], "building": True},
            "empty_result_gdf",
        ),
    ],
)
def test_osm_online_loader(
    area_gdf_fixture: str,
    query: OsmTagsFilter,
    expected_result_gdf_fixture: str,
    request: Any,
):
    """Test `OSMOnlineLoader.load()`."""
    _ = request.getfixturevalue("mock_osmnx")
    area_gdf = request.getfixturevalue(area_gdf_fixture)
    expected_result_gdf = request.getfixturevalue(expected_result_gdf_fixture)
    loader = OSMOnlineLoader()
    res = loader.load(area_gdf, query)
    assert "address" not in res.columns
    assert_frame_equal(res, expected_result_gdf, check_like=True)



"""Tests for OSM Loaders filters."""
import json
from contextlib import nullcontext as does_not_raise
from pathlib import Path
from typing import Any, Dict, List

import pytest
import requests_mock as r_mock
from requests import HTTPError

from srai.loaders.osm_loaders.filters import get_popular_tags


@pytest.fixture  # type: ignore
def popular_filter_api_data() -> Dict[str, Any]:
    """Load example taginfo API response data from file."""
    with (Path(__file__).parent / "popular_filter_example.json").open("rt") as f:
        res: Dict[str, Any] = json.load(f)
        return res


@pytest.fixture  # type: ignore
def expected_result_all() -> Dict[str, List[str]]:
    """Get expected results for whole api data."""
    return {
        "natural": ["wood"],
        "landuse": ["farmland", "residential"],
        "source": [
            "cadastre-dgi-fr source : Direction Générale des Impôts - Cadastre. Mise à jour : 2012",
            "bing",
        ],
        "highway": ["tertiary", "crossing"],
        "building:levels": ["2"],
        "source:date": ["2014-03-24"],
        "layer": ["1"],
    }


@pytest.fixture  # type: ignore
def expected_in_wiki_only() -> Dict[str, List[str]]:
    """Get expected results when using `in_wiki_only=True`."""
    return {
        "natural": ["wood"],
        "landuse": ["farmland", "residential"],
        "source": ["bing"],
        "highway": ["tertiary", "crossing"],
        "layer": ["1"],
    }


@pytest.fixture  # type: ignore
def expected_result_min_count_8m() -> Dict[str, List[str]]:
    """Get expected results when using `min_count=8_000_000`."""
    return {
        "natural": ["wood"],
        "landuse": ["farmland", "residential"],
    }


@pytest.fixture  # type: ignore
def expected_result_min_fraction() -> Dict[str, List[str]]:
    """Get expected results when using `min_fraction=0.001`."""
    return {
        "natural": ["wood"],
    }


@pytest.mark.parametrize(  # type: ignore
    "expected_result_fixture,status_code,in_wiki_only,min_count,min_fraction,expectation",
    [
        ("expected_result_all", 200, False, 0, 0.0, does_not_raise()),
        ("expected_in_wiki_only", 200, True, 0, 0.0, does_not_raise()),
        ("expected_result_min_count_8m", 200, False, 8_000_000, 0.0, does_not_raise()),
        ("expected_result_min_fraction", 200, False, 0, 0.001, does_not_raise()),
        ("expected_result_all", 400, False, 0, 0.0, pytest.raises(HTTPError)),
        ("expected_result_all", 500, False, 0, 0.0, pytest.raises(HTTPError)),
    ],
)
def test_get_popular_tags(
    expected_result_fixture: str,
    status_code: int,
    in_wiki_only: bool,
    min_count: int,
    min_fraction: float,
    expectation: Any,
    request: Any,
):
    """Test downloading popular tags from taginfo API."""
    requests_mock = request.getfixturevalue("requests_mock")
    api_response = request.getfixturevalue("popular_filter_api_data")
    expected_result = request.getfixturevalue(expected_result_fixture)

    with expectation:
        requests_mock.get(r_mock.ANY, json=api_response, status_code=status_code)
        result = get_popular_tags(
            in_wiki_only=in_wiki_only, min_count=min_count, min_fraction=min_fraction
        )
        assert result == expected_result
